# coding: utf-8
"""

"""

import datetime
import typing

from .. import db
from .users import User


class MarkdownImage(db.Model):
    __tablename__ = 'markdown_images'

    file_name = db.Column(db.Text, primary_key=True)
    content = db.Column(db.LargeBinary, nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey(User.id), nullable=True)
    utc_datetime = db.Column(db.DateTime, nullable=True)
    permanent = db.Column(db.Boolean, nullable=False, default=False)

    def __init__(self, file_name: str, content: bytes, user_id: int, utc_datetime: typing.Optional[datetime.datetime] = None, permanent: typing.Optional[bool] = False):
        self.file_name = file_name
        self.content = content
        self.user_id = user_id
        if utc_datetime is None:
            utc_datetime = datetime.datetime.utcnow()
        self.utc_datetime = utc_datetime
        self.permanent = permanent



# coding: utf-8
"""

"""


from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField
from wtforms.validators import InputRequired, Length


class EditGroupForm(FlaskForm):
    name = StringField(validators=[Length(min=1, max=100)])
    description = StringField()


class CreateGroupForm(FlaskForm):
    name = StringField(validators=[Length(min=1, max=100)])
    description = StringField()


class LeaveGroupForm(FlaskForm):
    pass


class DeleteGroupForm(FlaskForm):
    pass


class RemoveGroupMemberForm(FlaskForm):
    pass


class InviteUserForm(FlaskForm):
    user_id = IntegerField(validators=[InputRequired()])


class ToggleFavoriteActionForm(FlaskForm):
    action_id = IntegerField(validators=[InputRequired()])


class ToggleFavoriteInstrumentForm(FlaskForm):
    instrument_id = IntegerField(validators=[InputRequired()])


class NotificationModeForm(FlaskForm):
    pass


class OtherSettingsForm(FlaskForm):
    pass


class CreateAPITokenForm(FlaskForm):
    description = StringField('description', validators=[Length(min=1, max=100)])



# coding: utf-8
"""

"""

import enum
import datetime

from .. import db
from .objects import Objects

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


@enum.unique
class ObjectLogEntryType(enum.Enum):
    OTHER = 0
    CREATE_OBJECT = 1
    EDIT_OBJECT = 2
    RESTORE_OBJECT_VERSION = 3
    USE_OBJECT_IN_MEASUREMENT = 4
    POST_COMMENT = 5
    UPLOAD_FILE = 6
    USE_OBJECT_IN_SAMPLE_CREATION = 7
    CREATE_BATCH = 8
    ASSIGN_LOCATION = 9
    LINK_PUBLICATION = 10
    REFERENCE_OBJECT_IN_METADATA = 11


class ObjectLogEntry(db.Model):
    __tablename__ = 'object_log_entries'

    id = db.Column(db.Integer, primary_key=True)
    type = db.Column(db.Enum(ObjectLogEntryType), nullable=False)
    object_id = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    data = db.Column(db.JSON, nullable=False)
    utc_datetime = db.Column(db.DateTime, nullable=False)
    user = db.relationship('User')

    def __init__(self, type, object_id, user_id, data, utc_datetime=None):
        self.type = type
        self.object_id = object_id
        self.user_id = user_id
        self.data = data
        if utc_datetime is None:
            utc_datetime = datetime.datetime.utcnow()
        self.utc_datetime = utc_datetime

    def __repr__(self):
        return '<{0}(id={1.id}, type={1.type}, object_id={1.object_id}, user_id={1.user_id}, utc_datetime={1.utc_datetime}, data={1.data})>'.format(type(self).__name__, self)



# coding: utf-8
"""

"""

import sqlalchemy
import typing


def ascending(sorting_func: typing.Any) -> typing.Any:
    """
    Modify a sorting function to sort in ascending order.

    :param sorting_func: the original sorting function
    :return: the modified sorting function
    """
    def modified_sorting_func(current_columns, original_columns, sorting_func=sorting_func):
        return sqlalchemy.sql.asc(sorting_func(current_columns, original_columns))
    modified_sorting_func.require_original_columns = getattr(sorting_func, 'require_original_columns', False)
    return modified_sorting_func


def descending(sorting_func: typing.Any) -> typing.Any:
    """
    Modify a sorting function to sort in descending order.

    :param sorting_func: the original sorting function
    :return: the modified sorting function
    """
    def modified_sorting_func(current_columns, original_columns, sorting_func=sorting_func):
        return sqlalchemy.sql.desc(sorting_func(current_columns, original_columns))
    modified_sorting_func.require_original_columns = getattr(sorting_func, 'require_original_columns', False)
    return modified_sorting_func


def object_id() -> typing.Callable[[typing.Any], typing.Any]:
    """
    Create a sorting function to sort by object ID.

    :return: the sorting function
    """
    def sorting_func(current_columns: typing.Any, original_columns: typing.Any) -> typing.Any:
        return current_columns.object_id
    return sorting_func


def creation_date() -> typing.Callable[[typing.Any], typing.Any]:
    """
    Create a sorting function to sort by creation date.

    :return: the sorting function
    """
    def sorting_func(current_columns: typing.Any, original_columns: typing.Any) -> typing.Any:
        return sqlalchemy.sql.func.coalesce(original_columns.utc_datetime, current_columns.utc_datetime)
    sorting_func.require_original_columns = True
    return sorting_func


def last_modification_date() -> typing.Callable[[typing.Any], typing.Any]:
    """
    Create a sorting function to sort by last modification date.

    :return: the sorting function
    """
    def sorting_func(current_columns: typing.Any, original_columns: typing.Any) -> typing.Any:
        return current_columns.utc_datetime
    return sorting_func


def property_value(property_name: str) -> typing.Callable[[typing.Any], typing.Any]:
    """
    Create a sorting function to sort by an arbitrary property.

    :param property_name: the name of the property to sort by
    :return: the sorting function
    """
    def sorting_func(current_columns: typing.Any, original_columns: typing.Any) -> typing.Any:
        columns = current_columns
        return sqlalchemy.sql.expression.case([
            (
                columns.data[property_name]['_type'].astext == 'text',
                columns.data[property_name]['text'].astext
            ),
            (
                columns.data[property_name]['_type'].astext == 'quantity',
                sqlalchemy.func.to_char(columns.data[property_name]['magnitude_in_base_units'].astext.cast(sqlalchemy.Float), '00000000000000000000.00000000000000000000')
            ),
            (
                columns.data[property_name]['_type'].astext == 'bool',
                columns.data[property_name]['value'].astext
            ),
            (
                columns.data[property_name]['_type'].astext == 'datetime',
                columns.data[property_name]['utc_datetime'].astext
            ),
            (
                columns.data[property_name]['_type'].astext == 'sample',
                columns.data[property_name]['object_id'].astext
            )
        ], else_=sqlalchemy.sql.null())
    return sorting_func



# coding: utf-8
"""
Convert the instrument_log_entries table from containing content to using versioning.
"""

import os

MIGRATION_INDEX = 35
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    # Skip migration by condition
    column_names = db.session.execute("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'instrument_log_entries'
    """).fetchall()
    if ('content',) not in column_names:
        return False

    # Perform migration
    existing_log_entries = db.session.execute("""
        SELECT id, content, utc_datetime
        FROM instrument_log_entries
    """).fetchall()
    for log_entry_id, content, utc_datetime in existing_log_entries:
        db.session.execute("""
        INSERT INTO instrument_log_entry_versions (log_entry_id, version_id, content, utc_datetime)
        VALUES (:log_entry_id, 1, :content, :utc_datetime)
        """, {
            'log_entry_id': log_entry_id,
            'content': content,
            'utc_datetime': utc_datetime
        })
    existing_log_entry_categories = db.session.execute("""
    SELECT log_entry_id, category_id
    FROM instrument_log_entry_category_associations
    """).fetchall()
    for log_entry_id, category_id in existing_log_entry_categories:
        db.session.execute("""
        INSERT INTO instrument_log_entry_version_category_associations (log_entry_id, log_entry_version_id, category_id)
        VALUES (:log_entry_id, 1, :category_id)
        """, {
            'log_entry_id': log_entry_id,
            'category_id': category_id
        })
    db.session.execute("""
        DROP TABLE instrument_log_entry_category_associations
    """)
    db.session.execute("""
        ALTER TABLE instrument_log_entries
        DROP content
    """)
    db.session.execute("""
        ALTER TABLE instrument_log_entries
        DROP utc_datetime
    """)
    db.session.execute("""
        ALTER TABLE instrument_log_file_attachments
        ADD is_hidden BOOLEAN DEFAULT FALSE NOT NULL
    """)
    db.session.execute("""
        ALTER TABLE instrument_log_object_attachments
        ADD is_hidden BOOLEAN DEFAULT FALSE NOT NULL
    """)
    return True



# coding: utf-8
"""
Logic module for comments

Users with WRITE permissions can comment on samples or measurements. These
comments are immutable and therefore this module only allows the creation and
querying of comments.
"""

import typing

from .. import db
from . import user_log, object_log, objects, users, errors
from ..models import Comment


def create_comment(object_id: int, user_id: int, content: str) -> int:
    """
    Creates a new comment and adds it to the object and user logs.

    :param object_id: the ID of an existing object
    :param user_id: the ID of an existing user
    :param content: the text content for the new comment
    :return: the ID of the new comment
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    """
    # ensure that the object exists
    objects.get_object(object_id)
    # ensure that the user exists
    users.get_user(user_id)
    comment = Comment(
        object_id=object_id,
        user_id=user_id,
        content=content
    )
    db.session.add(comment)
    db.session.commit()
    object_log.post_comment(user_id=user_id, object_id=object_id, comment_id=comment.id)
    user_log.post_comment(user_id=user_id, object_id=object_id, comment_id=comment.id)
    return comment.id


def get_comments_for_object(object_id: int) -> typing.List[Comment]:
    """
    Returns a list of comments for an object.

    :param object_id: the ID of an existing object
    :return: the list of comments, sorted from oldest to newest
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    """
    comments = Comment.query.filter_by(object_id=object_id).order_by(db.asc(Comment.utc_datetime)).all()
    if not comments:
        # ensure that the object exists
        objects.get_object(object_id)
    return comments


def get_comment_for_object(object_id: int, comment_id: int) -> Comment:
    """
    Returns a specific comment for a given object.

    :param object_id: the ID of an existing object
    :param comment_id: the ID of an existing comment
    :return: the specified comment
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.CommentDoesNotExistError: when no comment with the given
        comment ID exists for this object
    """
    comment = Comment.query.filter_by(object_id=object_id, id=comment_id).first()
    if not comment:
        # ensure that the object exists
        objects.get_object(object_id)
        raise errors.CommentDoesNotExistError()
    return comment



# coding: utf-8
"""

"""

import collections
import typing

from .. import db
from .. import models
from . import errors
from . import object_log
from . import user_log
from .objects import get_object
from .users import get_user


class Publication(collections.namedtuple('Publication', ['doi', 'title', 'object_name'])):
    """
    This class provides an immutable wrapper around models.object_publications.ObjectPublication.
    """

    def __new__(cls, doi: str, title: str, object_name: str):
        self = super(Publication, cls).__new__(cls, doi, title, object_name)
        return self

    @classmethod
    def from_database(cls, publication: models.object_publications.ObjectPublication) -> 'Publication':
        return Publication(
            doi=publication.doi,
            title=publication.title,
            object_name=publication.object_name
        )


def link_publication_to_object(
        user_id: int,
        object_id: int,
        doi: str,
        title: typing.Optional[str] = None,
        object_name: typing.Optional[str] = None
):
    """
    Link a publication to an object.

    :param user_id: the ID of an existing user
    :param object_id: the ID of an existing object
    :param doi: the DOI of an existing publication
    :param title: the title of the publication
    :param object_name: the name of the object in the publication
    :raise errors.ObjectDoesNotExistError: when no object with the given object
        ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    """
    get_user(user_id)
    get_object(object_id)
    link = models.object_publications.ObjectPublication.query.filter_by(object_id=object_id, doi=doi).first()
    if link is None:
        link = models.object_publications.ObjectPublication(object_id=object_id, doi=doi)
    link.title = title
    link.object_name = object_name
    db.session.add(link)
    db.session.commit()
    user_log.link_publication(user_id=user_id, object_id=object_id, doi=doi, title=title, object_name=object_name)
    object_log.link_publication(user_id=user_id, object_id=object_id, doi=doi, title=title, object_name=object_name)


def get_publications_for_object(object_id: int) -> typing.Sequence[Publication]:
    """
    Get a list of publications linked to an object with a given ID.

    :param object_id: the ID of an existing object
    :return: a list of publications linked to the given object
    :raise errors.ObjectDoesNotExistError: when no object with the given object ID
        exists
    """
    get_object(object_id)
    links = models.object_publications.ObjectPublication.query.filter_by(object_id=object_id).all()
    return [
        Publication.from_database(link)
        for link in links
    ]


def get_object_ids_linked_to_doi(doi: str) -> typing.Sequence[int]:
    """
    Get a list of object IDs linked to a publication.

    :param doi: the DOI of the publication
    :return: the list of object IDs
    :raise errors.InvalidDOIError: when the DOI is invalid
    """
    doi = simplify_doi(doi)
    links = models.object_publications.ObjectPublication.query.filter_by(doi=doi).all()
    return [
        link.object_id
        for link in links
    ]


def simplify_doi(doi: str) -> str:
    """
    Simplify a digital object identifier (DOI).

    :param doi: the DOI to simplify
    :return: the simplified DOI
    :raise errors.InvalidDOIError: when the DOI is invalid
    """
    doi = doi.lower()
    doi_prefixes = [
        'doi:',
        'https://doi.org/'
    ]
    for doi_prefix in doi_prefixes:
        if doi.startswith(doi_prefix):
            doi = doi[len(doi_prefix):]
            break
    # DOI organizer part must start with '10.'
    if not doi.startswith('10.'):
        raise errors.InvalidDOIError()
    # DOI must be split into organizer and object parts by a slash
    if '/' not in doi[4:-1]:
        raise errors.InvalidDOIError()
    organizer_id, object_id = doi[3:].split('/', 1)
    if any(c not in '0123456789.' for c in organizer_id):
        raise errors.InvalidDOIError()
    if any(c in '<>"' for c in object_id):
        raise errors.InvalidDOIError()
    return doi



# coding: utf-8
"""
Implementation of validate_schema(schema)
"""


import datetime
import typing
import urllib.parse
import re

from ..errors import ValidationError
from .utils import units_are_valid
from .validate import validate

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def validate_schema(schema: dict, path: typing.Optional[typing.List[str]] = None) -> None:
    """
    Validates the given schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    if path is None:
        path = []
    if not isinstance(schema, dict):
        raise ValidationError('invalid schema (must be dict)', path)
    if 'type' not in schema:
        raise ValidationError('invalid schema (must contain type)', path)
    if not isinstance(schema['type'], str):
        raise ValidationError('invalid schema (type must be str)', path)
    if 'title' not in schema:
        raise ValidationError('invalid schema (must contain title)', path)
    if not isinstance(schema['title'], str):
        raise ValidationError('invalid schema (title must be str)', path)
    if path == [] and schema['type'] != 'object':
        raise ValidationError('invalid schema (root must be an object)', path)
    if schema['type'] == 'array':
        return _validate_array_schema(schema, path)
    elif schema['type'] == 'object':
        return _validate_object_schema(schema, path)
    elif schema['type'] == 'text':
        return _validate_text_schema(schema, path)
    elif schema['type'] == 'datetime':
        return _validate_datetime_schema(schema, path)
    elif schema['type'] == 'bool':
        return _validate_bool_schema(schema, path)
    elif schema['type'] == 'quantity':
        return _validate_quantity_schema(schema, path)
    elif schema['type'] == 'sample':
        return _validate_sample_schema(schema, path)
    elif schema['type'] == 'measurement':
        return _validate_measurement_schema(schema, path)
    elif schema['type'] == 'object_reference':
        return _validate_object_reference_schema(schema, path)
    elif schema['type'] == 'tags':
        return _validate_tags_schema(schema, path)
    elif schema['type'] == 'hazards':
        return _validate_hazards_schema(schema, path)
    elif schema['type'] == 'user':
        return _validate_user_schema(schema, path)
    else:
        raise ValidationError('invalid type', path)


def _validate_hazards_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validate the given GHS hazards schema and raise a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if path != ['hazards']:
        raise ValidationError('GHS hazards must be a top-level entry named "hazards"', path)
    if not isinstance(schema.get('note', ''), str):
        raise ValidationError('note must be a string', path)


def _validate_array_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given array schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'items', 'style', 'minItems', 'maxItems', 'default'}
    required_keys = {'type', 'title', 'items'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    has_min_items = False
    if 'minItems' in schema:
        if not isinstance(schema['minItems'], int):
            raise ValidationError('minItems must be an integer', path)
        elif schema['minItems'] < 0:
            raise ValidationError('minItems must not be negative', path)
        else:
            has_min_items = True
    has_max_items = False
    if 'maxItems' in schema:
        if not isinstance(schema['maxItems'], int):
            raise ValidationError('maxItems must be an integer', path)
        elif schema['maxItems'] < 0:
            raise ValidationError('maxItems must not be negative', path)
        else:
            has_max_items = True
    if has_min_items and has_max_items:
        if schema['minItems'] > schema['maxItems']:
            raise ValidationError('minItems must be less than or equal to maxItems', path)
    if 'style' in schema and schema['style'] not in ('table', 'list'):
        raise ValidationError('style must be either "list" or "table"', path)
    validate_schema(schema['items'], path + ['[?]'])
    if 'default' in schema:
        validate(schema['default'], schema, path + ['(default)'])


def _validate_tags_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given tags schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'default'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if 'default' in schema:
        validate({'_type': 'tags', 'tags': schema['default']}, schema, path + ['(default)'])
    if path != ['tags']:
        raise ValidationError('Tags must be a top-level entry named "tags"', path)


def _validate_object_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'properties', 'propertyOrder', 'required', 'default'}
    if not path:
        # the top level object may contain a list of properties to be displayed in a table of objects
        valid_keys.add('displayProperties')
        valid_keys.add('batch')
        valid_keys.add('batch_name_format')
        valid_keys.add('notebookTemplates')
    required_keys = {'type', 'title', 'properties'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)

    if not isinstance(schema['properties'], dict):
        raise ValidationError('properties must be dict', path)
    for property_name, property_schema in schema['properties'].items():
        if '__' in property_name:
            raise ValidationError('invalid property name: {}'.format(property_name), path)
        validate_schema(property_schema, path + [property_name])

    if 'required' in schema:
        if not isinstance(schema['required'], list):
            raise ValidationError('required must be list', path)
        for i, property_name in enumerate(schema['required']):
            if property_name not in schema['properties']:
                raise ValidationError('unknown required property: {}'.format(property_name), path)
            if property_name in schema['required'][:i]:
                raise ValidationError('duplicate required property: {}'.format(property_name), path)

    if 'hazards' in schema['properties'] and schema['properties']['hazards']['type'] == 'hazards' and 'hazards' not in schema.get('required', []):
        raise ValidationError('GHS hazards may not be optional', path)

    if not path:
        if 'name' not in schema['properties'] or schema['properties']['name']['type'] != 'text':
            raise ValidationError('Schema must include a text property "name"', path)
        if 'required' not in schema or 'name' not in schema['required']:
            raise ValidationError('"name" must be a required property for the root object', path)

    if 'propertyOrder' in schema:
        if not isinstance(schema['propertyOrder'], list):
            raise ValidationError('propertyOrder must be list', path)
        for i, property_name in enumerate(schema['propertyOrder']):
            if property_name not in schema['properties']:
                raise ValidationError('unknown propertyOrder property: {}'.format(property_name), path)
            if property_name in schema['propertyOrder'][:i]:
                raise ValidationError('duplicate propertyOrder property: {}'.format(property_name), path)

    if 'default' in schema:
        validate(schema['default'], schema)

    if 'displayProperties' in schema:
        if not isinstance(schema['displayProperties'], list):
            raise ValidationError('displayProperties must be list', path)
        for i, property_name in enumerate(schema['displayProperties']):
            if property_name not in schema['properties']:
                raise ValidationError('unknown display property: {}'.format(property_name), path)
            if property_name in schema['displayProperties'][:i]:
                raise ValidationError('duplicate displayProperties property: {}'.format(property_name), path)

    if 'batch' in schema:
        if not isinstance(schema['batch'], bool):
            raise ValidationError('batch must be bool', path)

    if 'batch_name_format' in schema:
        if not schema.get('batch', False):
            raise ValidationError('batch must be True for batch_name_format to be set', path)
        if not isinstance(schema['batch_name_format'], str):
            raise ValidationError('batch_name_format must be a string', path)
        try:
            schema['batch_name_format'].format(1)
        except (ValueError, KeyError):
            raise ValidationError('invalid batch_name_format', path)

    if 'notebookTemplates' in schema:
        _validate_notebook_templates(schema['notebookTemplates'])


def _validate_text_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given text object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'default', 'minLength', 'maxLength', 'choices', 'pattern', 'multiline', 'note', 'placeholder'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)

    if 'default' in schema and not isinstance(schema['default'], str):
        raise ValidationError('default must be str', path)
    if 'minLength' in schema and (not isinstance(schema['minLength'], int) or isinstance(schema['minLength'], bool)):
        raise ValidationError('minLength must be int', path)
    if 'maxLength' in schema and (not isinstance(schema['maxLength'], int) or isinstance(schema['maxLength'], bool)):
        raise ValidationError('maxLength must be int', path)
    if 'minLength' in schema and schema['minLength'] < 0:
        raise ValidationError('minLength must not be negative', path)
    if 'maxLength' in schema and schema['maxLength'] < 0:
        raise ValidationError('maxLength must not be negative', path)
    if 'minLength' in schema and 'maxLength' in schema and schema['maxLength'] < schema['minLength']:
        raise ValidationError('maxLength must not be less than minLength', path)
    if 'choices' in schema and not isinstance(schema['choices'], list):
        raise ValidationError('choices must be list', path)
    if 'choices' in schema and not schema['choices']:
        raise ValidationError('choices must not be empty', path)
    if 'choices' in schema:
        for i, choice in enumerate(schema['choices']):
            if not isinstance(choice, str):
                raise ValidationError('choice must be str', path + [str(i)])
            if choice.isspace():
                raise ValidationError('choice must contain more than whitespace', path + [str(i)])
    if 'placeholder' in schema and 'choices' in schema:
        raise ValidationError('placeholder cannot be used together with choices', path)
    if 'placeholder' in schema and not isinstance(schema['placeholder'], str):
        raise ValidationError('placeholder must be str', path)
    if 'pattern' in schema and not isinstance(schema['pattern'], str):
        raise ValidationError('pattern must be str', path)
    if 'pattern' in schema:
        try:
            re.compile(schema['pattern'])
        except re.error:
            raise ValidationError('pattern is no valid regular expression', path)
    if 'multiline' in schema and not isinstance(schema['multiline'], bool):
        raise ValidationError('multiline must be bool', path)

    if 'note' in schema and not isinstance(schema['note'], str):
        raise ValidationError('note must be str', path)


def _validate_datetime_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given datetime object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'default', 'note'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)

    if 'default' in schema:
        if not isinstance(schema['default'], str):
            raise ValidationError('default must be str', path)
        else:
            try:
                datetime.datetime.strptime(schema['default'], '%Y-%m-%d %H:%M:%S')
            except ValueError:
                raise ValidationError('invalid default value', path)

    if 'note' in schema and not isinstance(schema['note'], str):
        raise ValidationError('note must be str', path)


def _validate_bool_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given boolean object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'default', 'note'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)

    if 'default' in schema and not isinstance(schema['default'], bool):
        raise ValidationError('default must be bool', path)

    if 'note' in schema and not isinstance(schema['note'], str):
        raise ValidationError('note must be str', path)


def _validate_quantity_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given quantity object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'units', 'default', 'note', 'placeholder'}
    required_keys = {'type', 'title', 'units'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)

    if not isinstance(schema['units'], str):
        raise ValidationError('units must be str', path)
    elif not units_are_valid(schema['units']):
        raise ValidationError('invalid units', path)

    if 'default' in schema and not isinstance(schema['default'], float) and not isinstance(schema['default'], int):
        raise ValidationError('default must be float or int', path)

    if 'note' in schema and not isinstance(schema['note'], str):
        raise ValidationError('note must be str', path)

    if 'placeholder' in schema and not isinstance(schema['placeholder'], str):
        raise ValidationError('placeholder must be str', path)


def _validate_sample_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given sample object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if 'note' in schema and not isinstance(schema['note'], str):
        raise ValidationError('note must be str', path)


def _validate_measurement_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given measurement object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if 'note' in schema and not isinstance(schema['note'], str):
        raise ValidationError('note must be str', path)


def _validate_object_reference_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given object reference object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note', 'action_type_id', 'action_id'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if 'note' in schema and not isinstance(schema['note'], str):
        raise ValidationError('note must be str', path)
    if 'action_type_id' in schema and not isinstance(schema['action_type_id'], (int, type(None))):
        raise ValidationError('action_type_id must be int or None', path)
    if 'action_id' in schema and not isinstance(schema['action_id'], (int, type(None))):
        raise ValidationError('action_id must be int or None', path)


def _validate_notebook_templates(notebook_templates: typing.Any) -> None:
    """
    Validate the given notebook templates and raise a ValidationError if they are invalid.

    :param notebook_templates: the sampledb object schema
    :raise ValidationError: if the notebook templates are invalid
    """
    if not isinstance(notebook_templates, list):
        raise ValidationError('notebookTemplates must be a list', ['notebookTemplates'])

    for notebook_index, notebook_template in enumerate(notebook_templates):
        path = ['notebookTemplates', str(notebook_index)]
        if not isinstance(notebook_template, dict):
            raise ValidationError('notebook template must be a dict', path)
        valid_keys = {'title', 'url', 'params'}
        required_keys = valid_keys
        schema_keys = set(notebook_template.keys())
        invalid_keys = schema_keys - valid_keys
        if invalid_keys:
            raise ValidationError('unexpected keys in notebook template: {}'.format(invalid_keys), path)
        missing_keys = required_keys - schema_keys
        if missing_keys:
            raise ValidationError('missing keys in notebook template: {}'.format(missing_keys), path)
        if not isinstance(notebook_template['title'], str):
            raise ValidationError('notebook template title must be str', path)
        if not isinstance(notebook_template['url'], str):
            raise ValidationError('notebook template url must be str', path)
        if not notebook_template['url'].endswith('.ipynb'):
            raise ValidationError('notebook template url must end with .ipynb', path)
        base_url = 'https://iffjupyter.fz-juelich.de/templates/t/'
        test_url = urllib.parse.urljoin(base_url, notebook_template['url'])
        if not test_url.startswith(base_url) or urllib.parse.urlparse(base_url).netloc != urllib.parse.urlparse(test_url).netloc:
            raise ValidationError('notebook template url must be relative', path)
        if not isinstance(notebook_template['params'], dict):
            raise ValidationError('notebook template params must be a dict', path)
        for param_name, param_value in notebook_template['params'].items():
            if not isinstance(param_name, str):
                raise ValidationError('notebook template param names must be str', path)
            if not isinstance(param_value, str) and not isinstance(param_value, list):
                raise ValidationError('notebook template param values must be str or a list', path)
            if isinstance(param_value, list):
                for param_step in param_value:
                    if not isinstance(param_step, str) and not isinstance(param_step, int):
                        raise ValidationError('notebook template param value steps must be str or int', path)
            if isinstance(param_value, str):
                valid_param_values = {'object_id'}
                if param_value not in valid_param_values:
                    raise ValidationError('notebook template param value must be a list or one of {}'.format(valid_param_values), path)


def _validate_user_schema(schema: dict, path: typing.List[str]) -> None:
    """
    Validates the given user object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError('unexpected keys in schema: {}'.format(invalid_keys), path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError('missing keys in schema: {}'.format(missing_keys), path)
    if 'note' in schema and not isinstance(schema['note'], str):
        raise ValidationError('note must be str', path)



# coding: utf-8
"""

"""

from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, SelectField, FieldList, FormField, BooleanField
from wtforms.validators import Length, InputRequired

from ..logic.object_permissions import Permissions


class CreateProjectForm(FlaskForm):
    name = StringField(validators=[Length(min=1, max=100)])
    description = StringField()


class EditProjectForm(FlaskForm):
    name = StringField(validators=[Length(min=1, max=100)])
    description = StringField()


class LeaveProjectForm(FlaskForm):
    pass


class DeleteProjectForm(FlaskForm):
    pass


class RemoveProjectMemberForm(FlaskForm):
    pass


class RemoveProjectGroupForm(FlaskForm):
    pass


class OtherProjectIdForm(FlaskForm):
    project_id = IntegerField(
        validators=[InputRequired()]
    )
    add_user = BooleanField(default=False)


class InviteUserToProjectForm(FlaskForm):
    user_id = IntegerField(validators=[InputRequired()])
    other_project_ids = FieldList(FormField(OtherProjectIdForm), min_entries=0)


class InviteGroupToProjectForm(FlaskForm):
    group_id = IntegerField(validators=[InputRequired()])


class ProjectUserPermissionsForm(FlaskForm):
    user_id = IntegerField(
        validators=[InputRequired()]
    )
    permissions = SelectField(
        choices=[(p.name.lower(), p.name.lower()) for p in (Permissions.NONE, Permissions.READ, Permissions.WRITE, Permissions.GRANT)],
        validators=[InputRequired()]
    )


class ProjectGroupPermissionsForm(FlaskForm):
    group_id = IntegerField(
        validators=[InputRequired()]
    )
    permissions = SelectField(
        choices=[(p.name.lower(), p.name.lower()) for p in (Permissions.NONE, Permissions.READ, Permissions.WRITE, Permissions.GRANT)],
        validators=[InputRequired()]
    )


class ProjectPermissionsForm(FlaskForm):
    user_permissions = FieldList(FormField(ProjectUserPermissionsForm), min_entries=0)
    group_permissions = FieldList(FormField(ProjectGroupPermissionsForm), min_entries=0)


class AddSubprojectForm(FlaskForm):
    child_project_id = IntegerField(validators=[InputRequired()])
    child_can_add_users_to_parent = BooleanField(default=False)


class RemoveSubprojectForm(FlaskForm):
    child_project_id = IntegerField(validators=[InputRequired()])



"""
Module for generation sample/object labels as PDF files.
"""
import io
import base64
import os
from PIL import Image
from reportlab.pdfgen.canvas import Canvas
from reportlab.lib.pagesizes import A4, LETTER
from reportlab.lib.units import mm

import qrcode
import qrcode.image.pil


DEFAULT_PAPER_FORMAT = 'DIN A4 (Portrait)'
PAGE_SIZES = {
    'DIN A4 (Portrait)': A4,
    'DIN A4 (Landscape)': (A4[1], A4[0]),
    'Letter (Portrait)': LETTER,
    'Letter (Landscape)': (LETTER[1], LETTER[0])
}

HORIZONTAL_LABEL_MARGIN = 10
VERTICAL_LABEL_MARGIN = 10


def _generate_ghs_image_uris():
    ghs_image_uris = []
    GHS_IMAGE_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'static', 'img')
    for i in range(0, 10):
        ghs_image = Image.open(os.path.join(GHS_IMAGE_DIR, 'ghs0{}.png'.format(i))).convert('RGBA')
        ghs_background_image = Image.new('RGBA', ghs_image.size, (255, 255, 255, 255))
        ghs_image = Image.alpha_composite(ghs_background_image, ghs_image)
        image_stream = io.BytesIO()
        ghs_image.save(image_stream, format='png')
        image_stream.seek(0)
        ghs_image_uri = 'data:image/png;base64,' + base64.b64encode(image_stream.read()).decode('utf-8')
        ghs_image_uris.append(ghs_image_uri)
    return ghs_image_uris


GHS_IMAGE_URIS = _generate_ghs_image_uris()


def _draw_centered_wrapped_text(canvas, text, left_offset, width, top_cursor, font_name, font_size, line_height):
    lines = []
    while text:
        text = text.lstrip()
        line = text
        while canvas.stringWidth(line, font_name, font_size) > width:
            if ' ' in line:
                line = line.rsplit(' ', 1)[0]
            else:
                line = line[:-1]
        lines.append(line.strip())
        text = text[len(line):]

    canvas.setFont(font_name, font_size)
    for line in lines:
        if line:
            canvas.drawCentredString(left_offset + width / 2, top_cursor, line)
            top_cursor -= line_height * font_size
    return top_cursor


def _draw_label(canvas, sample_name, sample_creator, sample_creation_date, sample_id, ghs_classes, qrcode_uri, left_offset, top_offset, width, minimum_height, qrcode_width, ghs_classes_side_by_side=False, centered=True):
    font_name = "Helvetica"
    font_size = 8
    right_offset = left_offset + width
    top_cursor = top_offset - 3 * mm

    top_cursor = _draw_centered_wrapped_text(canvas, sample_name, left_offset + 0.5 * mm, width - 1 * mm, top_cursor, font_name + '-Bold', font_size, 1.2)
    canvas.line(left_offset, top_cursor + font_size / 2 * 1.2, right_offset, top_cursor + font_size / 2 * 1.2)
    top_cursor -= font_size / 2 * 1.2
    top_cursor = _draw_centered_wrapped_text(canvas, sample_creator, left_offset + 0.5 * mm, width - 1 * mm, top_cursor, font_name, font_size, 1.2)
    top_cursor = _draw_centered_wrapped_text(canvas, sample_creation_date, left_offset + 0.5 * mm, width - 1 * mm, top_cursor, font_name, font_size, 1.2)
    full_left_offset = left_offset
    full_right_offset = right_offset
    full_width = width

    width = qrcode_width
    if ghs_classes_side_by_side and ghs_classes:
        width += 20 * mm
    if centered:
        left_offset = left_offset / 2 + right_offset / 2 - width / 2
    right_offset = left_offset + width

    top_cursor_before_qrcode = top_cursor - font_size / 2 * 1.2
    canvas.drawImage(qrcode_uri, left_offset + (not ghs_classes_side_by_side) * (width / 2 - qrcode_width / 2), top_cursor - qrcode_width * 0.9, qrcode_width, qrcode_width)
    canvas.line(full_left_offset, top_cursor + font_size / 2 * 1.2, full_right_offset, top_cursor + font_size / 2 * 1.2)
    top_cursor -= font_size / 2 * 1.2
    top_cursor -= qrcode_width

    canvas.setFont("Helvetica", 6)
    if ghs_classes_side_by_side:
        canvas.drawCentredString(left_offset + qrcode_width / 2, top_cursor + font_size * 1.2, "#{}".format(sample_id))
    else:
        canvas.drawCentredString(left_offset / 2 + right_offset / 2, top_cursor + font_size * 1.2, "#{}".format(sample_id))
    if not ghs_classes_side_by_side:
        if ghs_classes:
            canvas.line(full_left_offset, top_cursor + font_size / 2 * 1.2, full_right_offset, top_cursor + font_size / 2 * 1.2)
            top_cursor -= font_size / 2 * 1.2
        else:
            top_cursor += font_size / 2 * 1.2
    top_cursor_after_qrcode = top_cursor

    if ghs_classes_side_by_side:
        top_cursor = top_cursor_before_qrcode
        left_offset += 20 * mm

    ghs_start_position = 0
    if len(ghs_classes) > 2:
        ghs_start_position = 2
        top_cursor -= 4.5 * mm
    if len(ghs_classes) == 1:
        top_cursor -= 9 * mm
        canvas.drawImage(GHS_IMAGE_URIS[ghs_classes[0]], left_offset + ((right_offset - left_offset) / 2 - 9 * mm / 2), top_cursor + 5, 9 * mm, 9 * mm, (255, 255, 255, 255, 255, 255))
    else:
        for i, ghs_class in enumerate(ghs_classes, start=ghs_start_position):
            if i % 3 == 0:
                top_cursor -= 9 * mm
            canvas.drawImage(GHS_IMAGE_URIS[ghs_class], left_offset + ((right_offset - left_offset) / 2 - 19 * mm / 2) + 0.5 * mm + (i % 3 == 1) * 9 * mm + (i % 3 == 2) * 4.5 * mm, top_cursor + 5 - (i % 3 == 2) * 4.5 * mm, 9 * mm, 9 * mm, (255, 255, 255, 255, 255, 255))
        if (len(ghs_classes) - 1) % 3 == 0:
            top_cursor -= 4.5 * mm

    if ghs_classes_side_by_side:
        left_offset -= 20 * mm
        top_cursor = min(top_cursor_after_qrcode + font_size / 2 * 1.2, top_cursor)

    height = top_offset - top_cursor
    if height < minimum_height:
        top_cursor -= minimum_height - height
        height = minimum_height

    if ghs_classes and ghs_classes_side_by_side:
        canvas.line(left_offset + width / 2, top_cursor_before_qrcode + font_size * 1.2, left_offset + width / 2, top_cursor)
    if not centered and full_width != width:
        canvas.line(left_offset + width, top_cursor_before_qrcode + font_size * 1.2, left_offset + width, top_cursor)
    left_offset = full_left_offset
    right_offset = full_right_offset
    width = full_width

    canvas.rect(left_offset, top_cursor, width, height, 1)
    return top_cursor


def _draw_long_label(canvas: Canvas, sample_name, sample_creator, sample_creation_date, sample_id, ghs_classes, qrcode_uri, left_offset, bottom_offset, minimum_width=0, include_qrcode=False, qrcode_size=15 * mm):
    font_name = "Helvetica"
    font_size = 8
    canvas.setFont(font_name, font_size)
    num_lines = 2
    if include_qrcode:
        upper_line_text1 = sample_name
        upper_line_text2 = " • #{}".format(sample_id)
        middle_line_text = sample_creator
        lower_line_text = sample_creation_date
        num_lines = 3
    else:
        upper_line_text1 = sample_name
        upper_line_text2 = " • #{}".format(sample_id)
        middle_line_text = sample_creator + " • " + sample_creation_date
        lower_line_text = ''

    min_height = 9 * mm
    if include_qrcode:
        min_height = max(min_height, qrcode_size - 2 * mm)
    padding = max(1 * mm, (min_height - num_lines * font_size) / 2)
    height = max(min_height, 2 * padding + num_lines * font_size)
    left_cursor = left_offset + 1 * mm
    canvas.setFont(font_name + '-Bold', font_size)
    canvas.drawString(left_cursor, bottom_offset + padding + (num_lines - 1) * font_size * 1.2, upper_line_text1)
    canvas.setFont(font_name, font_size)
    canvas.drawString(left_cursor + canvas.stringWidth(upper_line_text1, font_name + '-Bold', font_size), bottom_offset + padding + (num_lines - 1) * font_size * 1.2, upper_line_text2)
    canvas.drawString(left_cursor, bottom_offset + padding + (num_lines - 2) * font_size * 1.2, middle_line_text)
    if num_lines == 3:
        canvas.drawString(left_cursor, bottom_offset + padding, lower_line_text)
    text_width = max(
        canvas.stringWidth(upper_line_text1, font_name + '-Bold', font_size) + canvas.stringWidth(upper_line_text2, font_name, font_size),
        canvas.stringWidth(middle_line_text, font_name, font_size),
        canvas.stringWidth(lower_line_text, font_name, font_size)
    )
    left_cursor += 1 * mm + text_width
    for ghs_class in ghs_classes:
        canvas.drawImage(GHS_IMAGE_URIS[ghs_class], left_cursor, bottom_offset + height / 2 - 4.5 * mm, 9 * mm, 9 * mm, (255, 255, 255, 255, 255, 255))
        left_cursor += 9 * mm
    if height != 9 * mm:
        left_cursor += 1 * mm
    if include_qrcode:
        canvas.drawImage(qrcode_uri, left_cursor - 1 * mm, bottom_offset + height / 2 - qrcode_size / 2, qrcode_size, qrcode_size)
        left_cursor += qrcode_size - 2 * mm
    width = left_cursor - left_offset
    if width < minimum_width:
        width = minimum_width
    canvas.rect(left_offset, bottom_offset, width, height, 1)
    return bottom_offset


def create_labels(
        object_id, object_name, object_url, creation_user, creation_date, ghs_classes, paper_format=DEFAULT_PAPER_FORMAT,
        create_mixed_labels=True, create_long_labels=False, include_qrcode_in_long_labels=False,
        label_width=18, label_minimum_height=0, label_minimum_width=0, qrcode_width=18, ghs_classes_side_by_side=False,
        centered=True
):

    page_size = PAGE_SIZES.get(paper_format, PAGE_SIZES[DEFAULT_PAPER_FORMAT])
    page_width, page_height = page_size
    image = qrcode.make(object_url, image_factory=qrcode.image.pil.PilImage)
    image_stream = io.BytesIO()
    image.save(image_stream, format='png')
    image_stream.seek(0)
    qrcode_uri = 'data:image/png;base64,' + base64.b64encode(image_stream.read()).decode('utf-8')

    # 0 for unknown GHS classes with question mark icon
    ghs_classes = [
        ghs_class if 0 < ghs_class < 10 else 0
        for ghs_class in ghs_classes
    ]

    pdf_stream = io.BytesIO()
    canvas = Canvas(pdf_stream, pagesize=page_size)

    label_width = label_width * mm
    label_minimum_height = label_minimum_height * mm
    qrcode_width = qrcode_width * mm
    label_minimum_width = label_minimum_width * mm

    vertical_padding = 3 * mm
    horizontal_padding = 3 * mm
    horizontal_margin = HORIZONTAL_LABEL_MARGIN * mm
    vertical_margin = VERTICAL_LABEL_MARGIN * mm
    top_cursor = page_height - vertical_margin
    max_label_height = None
    while max_label_height is None or top_cursor - max_label_height > vertical_margin:
        if create_mixed_labels:
            bottom_cursor = min(
                _draw_long_label(canvas, object_name, creation_user, creation_date, object_id, ghs_classes, qrcode_uri, 10 * mm, top_cursor - 9 * mm),
                _draw_long_label(canvas, object_name, creation_user, creation_date, object_id, ghs_classes, qrcode_uri, 10 * mm, top_cursor - 27 * mm, include_qrcode=True),
                _draw_label(canvas, object_name, creation_user, creation_date, object_id, ghs_classes, qrcode_uri, 10 * mm, top_cursor - 32 * mm, 18 * mm, 52 * mm, 18 * mm),
                _draw_label(canvas, object_name, creation_user, creation_date, object_id, ghs_classes, qrcode_uri, 33 * mm, top_cursor - 32 * mm, 20 * mm, 52 * mm, 20 * mm),
                _draw_label(canvas, object_name, creation_user, creation_date, object_id, ghs_classes, qrcode_uri, 58 * mm, top_cursor - 32 * mm, 40 * mm, 35 * mm, 20 * mm, ghs_classes_side_by_side=True),
                _draw_label(canvas, object_name, creation_user, creation_date, object_id, ghs_classes, qrcode_uri, 103 * mm, top_cursor - 32 * mm, 75 * mm, 35 * mm, 20 * mm, ghs_classes_side_by_side=True, centered=False)
            )
        elif create_long_labels:
            bottom_cursor = _draw_long_label(canvas, object_name, creation_user, creation_date, object_id, ghs_classes, qrcode_uri, horizontal_margin, top_cursor - vertical_margin - (3.5 * mm if include_qrcode_in_long_labels else 0 * mm), minimum_width=label_minimum_width, include_qrcode=include_qrcode_in_long_labels)
        else:
            num_labels_per_row = int((page_width - 2 * horizontal_margin + horizontal_padding) / (label_width + horizontal_padding))
            horizontal_centering_offset = (page_width - 2 * horizontal_margin + horizontal_padding - num_labels_per_row * (label_width + horizontal_padding)) / 2
            if num_labels_per_row <= 0:
                # Better a cropped label than no label at all
                num_labels_per_row = 1
                horizontal_centering_offset = 0
            for i in range(num_labels_per_row):
                left_cursor = horizontal_margin + horizontal_centering_offset + i * (label_width + horizontal_padding)
                bottom_cursor = _draw_label(canvas, object_name, creation_user, creation_date, object_id, ghs_classes, qrcode_uri, left_cursor, top_cursor, label_width, label_minimum_height, qrcode_width, ghs_classes_side_by_side=ghs_classes_side_by_side, centered=centered)
        if max_label_height is None:
            max_label_height = top_cursor - bottom_cursor
        top_cursor = bottom_cursor - vertical_padding

    canvas.showPage()
    canvas.save()
    pdf_stream.seek(0)
    return pdf_stream.read()



# coding: utf-8
"""

"""

import json
import typing

import flask
import flask_login
from flask_wtf import FlaskForm
from wtforms import BooleanField, StringField, SelectField, IntegerField
from wtforms.validators import InputRequired, Length, ValidationError
import pygments
import pygments.lexers.data
import pygments.formatters

from . import frontend
from .. import models
from ..logic.actions import Action, create_action, get_action, get_actions, update_action, get_action_type
from ..logic.action_permissions import Permissions, action_is_public, get_user_action_permissions, set_action_public
from ..logic.favorites import get_user_favorite_action_ids
from ..logic.instruments import get_instrument, get_user_instruments
from ..logic.markdown_images import mark_referenced_markdown_images_as_permanent
from ..logic import errors, users
from ..logic.schemas.validate_schema import validate_schema
from ..logic.settings import get_user_settings
from .users.forms import ToggleFavoriteActionForm
from .utils import check_current_user_is_not_readonly, markdown_to_safe_html

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class ActionForm(FlaskForm):
    name = StringField(validators=[Length(min=1, max=100)])
    description = StringField()
    type = IntegerField()
    instrument = SelectField()
    schema = StringField(validators=[InputRequired()])
    is_public = BooleanField()
    is_user_specific = BooleanField(default=True)
    is_markdown = BooleanField(default=None)
    is_hidden = BooleanField(default=None)

    def validate_type(form, field):
        try:
            action_type_id = int(field.data)
        except ValueError:
            raise ValidationError("Invalid action type")
        try:
            action_type = get_action_type(action_type_id)
        except errors.ActionTypeDoesNotExistError:
            raise ValidationError("Unknown action type")
        if action_type.admin_only and not flask_login.current_user.is_admin:
            raise ValidationError("Actions with this type can only be created by administrators")


@frontend.route('/actions/')
@flask_login.login_required
def actions():
    action_type_id = flask.request.args.get('t', None)
    action_type = None
    if action_type_id is not None:
        try:
            action_type_id = int(action_type_id)
        except ValueError:
            # ensure old links still function
            action_type_id = {
                'samples': models.ActionType.SAMPLE_CREATION,
                'measurements': models.ActionType.MEASUREMENT,
                'simulations': models.ActionType.SIMULATION
            }.get(action_type_id, None)
        try:
            action_type = get_action_type(action_type_id)
        except errors.ActionTypeDoesNotExistError:
            # fall back to all objects
            action_type_id = None
    user_id = flask.request.args.get('user_id', None)
    if user_id is not None:
        try:
            user_id = int(user_id)
        except ValueError:
            user_id = None
            flask.flash('Invalid user ID.', 'error')
    if user_id is not None:
        try:
            users.get_user(user_id)
        except errors.UserDoesNotExistError:
            user_id = None
            flask.flash('Invalid user ID.', 'error')
    actions = []
    action_permissions = {}
    for action in get_actions(action_type_id=action_type_id):
        if user_id is not None and action.user_id != user_id:
            continue
        if action.is_hidden and not flask_login.current_user.is_admin and user_id != flask_login.current_user.id:
            continue
        permissions = get_user_action_permissions(user_id=flask_login.current_user.id, action_id=action.id)
        if Permissions.READ in permissions:
            actions.append(action)
            action_permissions[action.id] = permissions
    user_favorite_action_ids = get_user_favorite_action_ids(flask_login.current_user.id)
    # Sort by: favorite / not favorite, instrument name (independent actions first), action name
    actions.sort(key=lambda action: (
        0 if action.id in user_favorite_action_ids else 1,
        action.user.name.lower() if action.user else '',
        action.instrument.name.lower() if action.instrument else '',
        action.name.lower()
    ))
    toggle_favorite_action_form = ToggleFavoriteActionForm()
    return flask.render_template(
        'actions/actions.html',
        actions=actions,
        action_type=action_type,
        action_permissions=action_permissions, Permissions=Permissions,
        user_favorite_action_ids=user_favorite_action_ids,
        toggle_favorite_action_form=toggle_favorite_action_form
    )


@frontend.route('/actions/<int:action_id>', methods=['GET', 'POST'])
@flask_login.login_required
def action(action_id):
    try:
        action = get_action(action_id)
    except errors.ActionDoesNotExistError:
        return flask.abort(404)
    permissions = get_user_action_permissions(action_id, flask_login.current_user.id)
    if Permissions.READ not in permissions:
        return flask.abort(403)
    may_edit = Permissions.WRITE in permissions
    mode = flask.request.args.get('mode', None)
    if mode == 'edit':
        check_current_user_is_not_readonly()
        if not may_edit:
            return flask.abort(403)
        return show_action_form(action)
    return flask.render_template(
        'actions/action.html',
        action=action,
        may_edit=may_edit,
        is_public=action_is_public(action_id)
    )


@frontend.route('/actions/new/', methods=['GET', 'POST'])
@flask_login.login_required
def new_action():
    check_current_user_is_not_readonly()
    action_type_id = flask.request.args.get('action_type_id')
    if action_type_id is not None:
        try:
            action_type_id = int(action_type_id)
        except ValueError:
            action_type_id = None
    previous_action = None
    previous_action_id = flask.request.args.get('previous_action_id', None)
    if previous_action_id is not None:
        try:
            previous_action_id = int(previous_action_id)
        except ValueError:
            previous_action_id = None
    if previous_action_id:
        try:
            permissions = get_user_action_permissions(previous_action_id, flask_login.current_user.id)
            if Permissions.READ in permissions:
                previous_action = get_action(previous_action_id)
            else:
                flask.flash('Insufficient permissions to use the requested previous action.', 'error')
                previous_action = None
        except errors.ActionDoesNotExistError:
            flask.flash('The requested previous action does not exist.', 'error')
        else:
            if previous_action.type.admin_only and not flask_login.current_user.is_admin:
                flask.flash('Only administrators can create actions of this type.', 'error')
                return flask.redirect(flask.url_for('.action', action_id=previous_action_id))
    return show_action_form(None, previous_action, action_type_id)


def _get_lines_for_path(schema: dict, path: typing.List[str]) -> typing.Optional[typing.Set[int]]:
    schema_entry = schema
    parent = None
    key = None
    for property in path:
        if property == '[?]' and 'items' in schema_entry and isinstance(schema_entry['items'], dict):
            parent = schema_entry
            key = 'items'
            schema_entry = parent[key]
        elif 'properties' in schema_entry and property in schema_entry['properties'] and isinstance(schema_entry['properties'][property], dict):
            parent = schema_entry['properties']
            key = property
            schema_entry = parent[key]
        else:
            break
    if parent is None or key is None:
        return None

    parent[key] = {"INTERNAL_MARKER_START": None, "INTERNAL_MARKER_CONTENT": parent[key], "INTERNAL_MARKER_END": None}

    schema_json = json.dumps(schema, indent=2)
    in_error = False
    error_lines = []
    skip_lines = []
    schema_json_lines = schema_json.splitlines(keepends=False)
    for i, line in enumerate(schema_json_lines):
        if line.endswith('"INTERNAL_MARKER_START": null,'):
            skip_lines.append(i)
            skip_lines.append(i + 1)
            in_error = True
        if in_error:
            error_lines.append(i)
        if line.endswith('"INTERNAL_MARKER_END": null'):
            skip_lines.append(i - 1)
            skip_lines.append(i)
            in_error = False

    new_error_lines = set()
    for i in reversed(error_lines):
        if i in skip_lines:
            new_error_lines = {i - 1 for i in new_error_lines}
        else:
            new_error_lines.add(i + 1)
            new_error_lines.add(i)
            new_error_lines.add(i - 1)
    return new_error_lines


def show_action_form(action: typing.Optional[Action] = None, previous_action: typing.Optional[Action] = None, action_type_id: typing.Optional[int] = None):
    if action is not None:
        schema_json = json.dumps(action.schema, indent=2)
        submit_text = "Save"
    elif previous_action is not None:
        schema_json = json.dumps(previous_action.schema, indent=2)
        submit_text = "Create"
    else:
        schema_json = json.dumps({
            'title': '',
            'type': 'object',
            'properties': {
                'name': {
                    'title': 'Name',
                    'type': 'text'
                }
            },
            'required': ['name']
        }, indent=2)
        submit_text = "Create"
    may_change_public = action is None or action.user_id is not None
    may_set_user_specific = action is None and flask_login.current_user.is_admin
    schema = None
    pygments_output = None
    error_message = None
    action_form = ActionForm()
    if action is not None:
        if action.instrument_id:
            action_form.instrument.choices = [
                (str(action.instrument_id), get_instrument(action.instrument_id).name)
            ]
            action_form.instrument.data = str(action.instrument_id)
        else:
            action_form.instrument.choices = [('-1', '-')]
            action_form.instrument.data = str(-1)
        action_form.type.data = action.type.id
        if action_form.name.data is None:
            action_form.is_public.data = None
    else:
        user_instrument_ids = get_user_instruments(flask_login.current_user.id, exclude_hidden=True)
        action_form.instrument.choices = [('-1', '-')] + [
            (str(instrument_id), get_instrument(instrument_id).name)
            for instrument_id in user_instrument_ids
        ]
        if action_form.instrument.data is None or action_form.instrument.data == str(None):
            if previous_action is not None and previous_action.instrument_id in user_instrument_ids:
                action_form.instrument.data = str(previous_action.instrument_id)
            else:
                action_form.instrument.data = '-1'

    form_is_valid = False
    if not action_form.is_submitted():
        action_form.type.data = action_type_id
    elif action_form.validate_on_submit():
        form_is_valid = True
        action_type_id = action_form.type.data
    else:
        action_type_id = None

    if action is not None:
        if action_form.name.data is None:
            action_form.name.data = action.name
        if action_form.description.data is None:
            action_form.description.data = action.description
        if not action_form.is_submitted():
            action_form.is_markdown.data = (action.description_as_html is not None)
            action_form.is_hidden.data = action.is_hidden
        if action_form.type.data is None:
            action_form.type.data = action.type.id
        if action_form.is_public.data is None:
            action_form.is_public.data = action_is_public(action.id)
    elif previous_action is not None:
        if action_form.name.data is None:
            action_form.name.data = previous_action.name
        if action_form.description.data is None:
            action_form.description.data = previous_action.description
        if not action_form.is_submitted():
            action_form.is_markdown.data = (previous_action.description_as_html is not None)
            action_form.is_hidden.data = False
        if action_form.type.data is None:
            action_form.type.data = previous_action.type.id
        if action_form.is_public.data is None:
            action_form.is_public.data = action_is_public(previous_action.id)

    if action_form.schema.data:
        schema_json = action_form.schema.data
    else:
        action_form.schema.data = schema_json
    if schema_json:
        all_lines = set(range(1, 1 + len(schema_json.splitlines())))
        error_lines = set()
        if 'INTERNAL_MARKER' in schema_json:
            error_lines = all_lines
            error_message = "Failed to parse schema: Must not contain INTERNAL_MARKER"
        else:
            try:
                schema = json.loads(schema_json)
            except json.JSONDecodeError as e:
                error_lines = {e.lineno}
                error_message = "Failed to parse as JSON: {}".format(e.msg)
            except Exception as e:
                error_lines = all_lines
                error_message = "Failed to parse as JSON: {}".format(str(e))
        if schema is not None:
            try:
                validate_schema(schema)
            except errors.ValidationError as e:
                error_message = e.message
                if not e.paths:
                    error_lines = all_lines
                else:
                    schema_json = json.dumps(schema, indent=2)
                    all_lines = set(range(1, 1 + len(schema_json)))
                    for path in e.paths:
                        new_error_lines = _get_lines_for_path(schema, path)
                        if new_error_lines is None:
                            error_lines = all_lines
                            break
                        else:
                            error_lines.update(new_error_lines)
                    else:
                        error_lines = {i + 1 for i in error_lines}
            except Exception as e:
                error_lines = all_lines
                error_message = "Unknown errror: {}".format(str(e))
        lexer = pygments.lexers.data.JsonLexer()
        formatter = pygments.formatters.HtmlFormatter(cssclass='pygments', hl_lines=list(error_lines))
        pygments_output = pygments.highlight(schema_json, lexer, formatter)
    if schema is not None and error_message is None and form_is_valid:
        name = action_form.name.data
        description = action_form.description.data
        if action_form.is_markdown.data:
            description_as_html = markdown_to_safe_html(description)
            mark_referenced_markdown_images_as_permanent(description_as_html)
        else:
            description_as_html = None

        instrument_id = action_form.instrument.data
        is_public = action_form.is_public.data
        is_hidden = action_form.is_hidden.data
        try:
            instrument_id = int(instrument_id)
        except ValueError:
            instrument_id = None
        if instrument_id < 0:
            instrument_id = None
        if action is None:
            if action_form.is_user_specific.data or not may_set_user_specific:
                user_id = flask_login.current_user.id
            else:
                user_id = None
            action = create_action(
                action_type_id,
                name,
                description,
                schema,
                instrument_id,
                user_id,
                description_as_html=description_as_html,
                is_hidden=is_hidden
            )
            flask.flash('The action was created successfully.', 'success')
            if may_change_public and is_public:
                set_action_public(action.id, True)
        else:
            update_action(
                action.id,
                name,
                description,
                schema,
                description_as_html=description_as_html,
                is_hidden=is_hidden
            )
            flask.flash('The action was updated successfully.', 'success')
            if may_change_public and is_public is not None:
                set_action_public(action.id, is_public)
        return flask.redirect(flask.url_for('.action', action_id=action.id))

    sample_action_type = get_action_type(models.ActionType.SAMPLE_CREATION)
    measurement_action_type = get_action_type(models.ActionType.MEASUREMENT)

    use_schema_editor = get_user_settings(flask_login.current_user.id)["USE_SCHEMA_EDITOR"]
    return flask.render_template(
        'actions/action_form.html',
        action_form=action_form,
        pygments_output=pygments_output,
        error_message=error_message,
        schema_json=schema_json,
        submit_text=submit_text,
        sample_action_type=sample_action_type,
        measurement_action_type=measurement_action_type,
        use_schema_editor=use_schema_editor,
        may_change_type=action is None,
        may_change_instrument=action is None,
        may_set_user_specific=may_set_user_specific,
        may_change_public=may_change_public
    )



# coding: utf-8
"""

"""

import json
import os

import itsdangerous
import flask
import flask_login
from flask_wtf import FlaskForm
from wtforms import StringField, SelectMultipleField, BooleanField, MultipleFileField, IntegerField
from wtforms.validators import Length, DataRequired, ValidationError

from . import frontend
from ..logic.instruments import get_instruments, get_instrument, create_instrument, update_instrument, set_instrument_responsible_users
from ..logic.instrument_log_entries import get_instrument_log_entries, create_instrument_log_entry, get_instrument_log_file_attachment, create_instrument_log_file_attachment, create_instrument_log_object_attachment, get_instrument_log_object_attachments, get_instrument_log_categories, InstrumentLogCategoryTheme, create_instrument_log_category, update_instrument_log_category, delete_instrument_log_category, update_instrument_log_entry, hide_instrument_log_file_attachment, hide_instrument_log_object_attachment, get_instrument_log_entry, get_instrument_log_object_attachment
from ..logic.actions import ActionType
from ..logic.errors import InstrumentDoesNotExistError, InstrumentLogFileAttachmentDoesNotExistError, ObjectDoesNotExistError, UserDoesNotExistError, InstrumentLogEntryDoesNotExistError, InstrumentLogObjectAttachmentDoesNotExistError
from ..logic.favorites import get_user_favorite_instrument_ids
from ..logic.markdown_images import mark_referenced_markdown_images_as_permanent
from ..logic.users import get_users, get_user
from ..logic.objects import get_object
from ..logic.object_permissions import Permissions, get_object_info_with_permissions
from ..logic.settings import get_user_settings, set_user_settings
from .users.forms import ToggleFavoriteInstrumentForm
from .utils import check_current_user_is_not_readonly, markdown_to_safe_html, generate_qrcode
from .validators import MultipleObjectIdValidator


__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class MultipleIntegerField(SelectMultipleField):
    def pre_validate(self, form):
        if self.data:
            for d in self.data:
                try:
                    int(d)
                except ValueError:
                    raise ValidationError("Invalid value")


class SelectMultipleFieldFix(SelectMultipleField):
    def pre_validate(self, form):
        # TODO: use SelectMultipleField directly after wtforms release
        # SelectMultipleField does not yet support the validate_choice flag
        # used by SelectField. This is already fixed in master of wtforms, but
        # until the new version is released, this adds support for the flag.
        if not self.validate_choice:
            return None
        super(SelectMultipleFieldFix, self).pre_validate(form)


class InstrumentLogEntryForm(FlaskForm):
    content = StringField()
    files = MultipleFileField()
    objects = SelectMultipleFieldFix(validators=[MultipleObjectIdValidator(Permissions.READ)], validate_choice=False)
    categories = SelectMultipleField()
    log_entry_id = IntegerField()
    existing_files = MultipleIntegerField()
    existing_objects = MultipleIntegerField()


class InstrumentLogOrderForm(FlaskForm):
    ascending = BooleanField()


@frontend.route('/instruments/')
@flask_login.login_required
def instruments():
    all_instruments = get_instruments()
    instruments = []
    for instrument in all_instruments:
        if instrument.is_hidden and not flask_login.current_user.is_admin and flask_login.current_user not in instrument.responsible_users:
            continue
        instruments.append(instrument)
    user_favorite_instrument_ids = get_user_favorite_instrument_ids(flask_login.current_user.id)
    # Sort by: favorite / not favorite, instrument name
    instruments.sort(key=lambda instrument: (
        0 if instrument.id in user_favorite_instrument_ids else 1,
        instrument.name.lower()
    ))
    toggle_favorite_instrument_form = ToggleFavoriteInstrumentForm()
    return flask.render_template(
        'instruments/instruments.html',
        instruments=instruments,
        user_favorite_instrument_ids=user_favorite_instrument_ids,
        toggle_favorite_instrument_form=toggle_favorite_instrument_form
    )


@frontend.route('/instruments/<int:instrument_id>', methods=['GET', 'POST'])
@flask_login.login_required
def instrument(instrument_id):
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)
    is_instrument_responsible_user = any(
        responsible_user.id == flask_login.current_user.id
        for responsible_user in instrument.responsible_users
    )
    if is_instrument_responsible_user or instrument.users_can_view_log_entries:
        instrument_log_entries = get_instrument_log_entries(instrument_id)
        attached_object_ids = set()
        for log_entry in instrument_log_entries:
            for object_attachment in get_instrument_log_object_attachments(log_entry.id):
                attached_object_ids.add(object_attachment.object_id)
        attached_object_infos = get_object_info_with_permissions(
            user_id=flask_login.current_user.id,
            permissions=Permissions.READ,
            object_ids=list(attached_object_ids)
        )
        attached_object_names = {
            object_info.object_id: object_info.name_text
            for object_info in attached_object_infos
        }
    else:
        instrument_log_entries = None
        attached_object_names = {}
    if is_instrument_responsible_user or instrument.users_can_create_log_entries:
        serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='instrument-log-mobile-upload')
        token = serializer.dumps([flask_login.current_user.id, instrument_id])
        mobile_upload_url = flask.url_for('.instrument_log_mobile_file_upload', instrument_id=instrument_id, token=token, _external=True)
        mobile_upload_qrcode = generate_qrcode(mobile_upload_url, should_cache=False)
    else:
        mobile_upload_url = None
        mobile_upload_qrcode = None
    instrument_log_entry_form = InstrumentLogEntryForm()
    if flask.current_app.config["LOAD_OBJECTS_IN_BACKGROUND"]:
        instrument_log_entry_form.objects.choices = []
    else:
        instrument_log_entry_form.objects.choices = [
            (str(object_info.object_id), "{} (#{})".format(object_info.name_text, object_info.object_id))
            for object_info in get_object_info_with_permissions(user_id=flask_login.current_user.id, permissions=Permissions.READ)
        ]
    instrument_log_categories = get_instrument_log_categories(instrument_id)
    instrument_log_entry_form.categories.choices = [
        (str(category.id), category.title)
        for category in instrument_log_categories
    ]
    if instrument_log_entry_form.validate_on_submit():
        check_current_user_is_not_readonly()
        if 'action_edit_content' in flask.request.form:
            if is_instrument_responsible_user or instrument.users_can_view_log_entries:
                try:
                    log_entry_id = int(instrument_log_entry_form.log_entry_id.data)
                    log_entry = get_instrument_log_entry(log_entry_id)
                except ValueError:
                    return flask.abort(400)
                except InstrumentLogEntryDoesNotExistError:
                    return flask.abort(400)
                update_instrument_log_entry(
                    log_entry_id=log_entry_id,
                    content=instrument_log_entry_form.content.data,
                    category_ids=[
                        int(category_id) for category_id in instrument_log_entry_form.categories.data
                    ]
                )
                for file_storage in instrument_log_entry_form.files.data:
                    if file_storage.filename:
                        file_name = file_storage.filename
                        content = file_storage.stream.read()
                        create_instrument_log_file_attachment(log_entry.id, file_name, content)
                previously_attached_object_ids = {
                    object_attachment.object_id: object_attachment
                    for object_attachment in log_entry.object_attachments
                }
                for object_id in instrument_log_entry_form.objects.data:
                    if int(object_id) in previously_attached_object_ids:
                        object_attachment = previously_attached_object_ids[int(object_id)]
                        if object_attachment.is_hidden:
                            hide_instrument_log_object_attachment(object_attachment.id, set_hidden=False)
                        continue
                    try:
                        create_instrument_log_object_attachment(log_entry.id, int(object_id))
                    except ObjectDoesNotExistError:
                        continue
                if instrument_log_entry_form.existing_files.data:
                    for file_attachment_id in instrument_log_entry_form.existing_files.data:
                        file_attachment_id = int(file_attachment_id)
                        try:
                            file_attachment = get_instrument_log_file_attachment(file_attachment_id)
                            if file_attachment.log_entry_id == log_entry_id:
                                hide_instrument_log_file_attachment(file_attachment_id)
                        except InstrumentLogFileAttachmentDoesNotExistError:
                            continue
                if instrument_log_entry_form.existing_objects.data:
                    for object_attachment_id in instrument_log_entry_form.existing_objects.data:
                        object_attachment_id = int(object_attachment_id)
                        try:
                            object_attachment = get_instrument_log_object_attachment(object_attachment_id)
                            if object_attachment.log_entry_id == log_entry_id:
                                hide_instrument_log_object_attachment(object_attachment_id)
                        except InstrumentLogObjectAttachmentDoesNotExistError:
                            continue
                flask.flash('You have successfully updated the instrument log entry.', 'success')
                return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
            else:
                return flask.abort(400)
        elif is_instrument_responsible_user or instrument.users_can_create_log_entries:
            log_entry_content = instrument_log_entry_form.content.data
            log_entry_has_files = any(file_storage.filename for file_storage in instrument_log_entry_form.files.data)
            object_ids = []
            for object_id in instrument_log_entry_form.objects.data:
                try:
                    object_ids.append(get_object(int(object_id)).id)
                except ObjectDoesNotExistError:
                    continue
                except ValueError:
                    continue
            log_entry_has_objects = bool(object_ids)
            if log_entry_content or log_entry_has_files or log_entry_has_objects:
                log_entry = create_instrument_log_entry(
                    instrument_id=instrument_id,
                    user_id=flask_login.current_user.id,
                    content=instrument_log_entry_form.content.data,
                    category_ids=[
                        int(category_id) for category_id in instrument_log_entry_form.categories.data
                    ]
                )
                for file_storage in instrument_log_entry_form.files.data:
                    if file_storage.filename:
                        file_name = file_storage.filename
                        content = file_storage.stream.read()
                        create_instrument_log_file_attachment(log_entry.id, file_name, content)
                for object_id in instrument_log_entry_form.objects.data:
                    try:
                        create_instrument_log_object_attachment(log_entry.id, int(object_id))
                    except ObjectDoesNotExistError:
                        continue
                flask.flash('You have created a new instrument log entry.', 'success')
                return flask.redirect(flask.url_for(
                    '.instrument',
                    instrument_id=instrument_id,
                    _anchor=f'log_entry-{log_entry.id}'
                ))
            else:
                flask.flash('Please enter a log entry text, upload a file or select an object to create a log entry.', 'error')
                return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
        else:
            flask.flash('You cannot create a log entry for this instrument.', 'error')
            return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
    instrument_log_order_ascending = get_user_settings(flask_login.current_user.id)["INSTRUMENT_LOG_ORDER_ASCENDING"]
    if instrument_log_entries is not None and not instrument_log_order_ascending:
        instrument_log_entries.reverse()
    return flask.render_template(
        'instruments/instrument.html',
        instrument=instrument,
        instrument_log_entries=instrument_log_entries,
        instrument_log_categories=instrument_log_categories,
        attached_object_names=attached_object_names,
        is_instrument_responsible_user=is_instrument_responsible_user,
        instrument_log_entry_form=instrument_log_entry_form,
        instrument_log_order_form=InstrumentLogOrderForm(),
        mobile_upload_url=mobile_upload_url,
        mobile_upload_qrcode=mobile_upload_qrcode,
        instrument_log_order_ascending=instrument_log_order_ascending,
        ActionType=ActionType
    )


class InstrumentForm(FlaskForm):
    name = StringField(validators=[Length(min=1, max=100)])
    description = StringField()
    instrument_responsible_users = SelectMultipleField()
    is_markdown = BooleanField(default=None)
    notes = StringField()
    notes_are_markdown = BooleanField(default=None)
    users_can_create_log_entries = BooleanField(default=False)
    users_can_view_log_entries = BooleanField(default=False)
    categories = StringField(validators=[DataRequired()])
    create_log_entry_default = BooleanField(default=False)
    is_hidden = BooleanField(default=False)


@frontend.route('/instruments/new', methods=['GET', 'POST'])
@flask_login.login_required
def new_instrument():
    if not flask_login.current_user.is_admin:
        return flask.abort(403)
    check_current_user_is_not_readonly()
    instrument_form = InstrumentForm()
    instrument_form.instrument_responsible_users.choices = [
        (str(user.id), user.name)
        for user in get_users()
    ]
    if instrument_form.validate_on_submit():
        if instrument_form.is_markdown.data:
            description_as_html = markdown_to_safe_html(instrument_form.description.data)
            mark_referenced_markdown_images_as_permanent(description_as_html)
        else:
            description_as_html = None

        if instrument_form.notes_are_markdown.data:
            notes_as_html = markdown_to_safe_html(instrument_form.notes.data)
            mark_referenced_markdown_images_as_permanent(notes_as_html)
        else:
            notes_as_html = None

        instrument = create_instrument(
            instrument_form.name.data,
            instrument_form.description.data,
            description_as_html=description_as_html,
            notes=instrument_form.notes.data,
            notes_as_html=notes_as_html,
            users_can_create_log_entries=instrument_form.users_can_create_log_entries.data,
            users_can_view_log_entries=instrument_form.users_can_view_log_entries.data,
            create_log_entry_default=instrument_form.create_log_entry_default.data,
            is_hidden=instrument_form.is_hidden.data
        )
        flask.flash('The instrument was created successfully.', 'success')
        instrument_responsible_user_ids = [
            int(user_id)
            for user_id in instrument_form.instrument_responsible_users.data
        ]
        set_instrument_responsible_users(instrument.id, instrument_responsible_user_ids)
        try:
            category_data = json.loads(instrument_form.categories.data)
        except Exception:
            pass
        else:
            themes_by_name = {
                theme.name.lower(): theme
                for theme in list(InstrumentLogCategoryTheme)
            }
            category_keys = {'id', 'title', 'theme'}
            if not isinstance(category_data, list):
                category_data = ()
            for category in category_data:
                # skip any invalid entries
                if not isinstance(category, dict):
                    continue
                if set(category.keys()) != category_keys:
                    continue
                if not all(isinstance(category[key], str) for key in category_keys):
                    continue
                category_title = category['title'].strip()
                if not category_title:
                    continue
                if len(category_title) > 100:
                    continue
                try:
                    category_id = int(category['id'])
                except ValueError:
                    continue
                category_theme = themes_by_name.get(category['theme'], InstrumentLogCategoryTheme.GRAY)
                if category_id < 0:
                    create_instrument_log_category(
                        instrument.id,
                        category['title'],
                        category_theme
                    )
        return flask.redirect(flask.url_for('.instrument', instrument_id=instrument.id))
    return flask.render_template(
        'instruments/instrument_form.html',
        submit_text='Create Instrument',
        instrument_log_category_themes=sorted(InstrumentLogCategoryTheme, key=lambda t: t.value),
        instrument_form=instrument_form
    )


@frontend.route('/instruments/<int:instrument_id>/edit', methods=['GET', 'POST'])
@flask_login.login_required
def edit_instrument(instrument_id):
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)
    check_current_user_is_not_readonly()
    if not flask_login.current_user.is_admin and flask_login.current_user not in instrument.responsible_users:
        return flask.abort(403)
    instrument_form = InstrumentForm()
    instrument_form.name.default = instrument.name
    instrument_form.description.default = instrument.description
    instrument_form.notes.default = instrument.notes
    instrument_form.instrument_responsible_users.choices = [
        (str(user.id), user.name)
        for user in get_users()
    ]
    instrument_form.instrument_responsible_users.default = [
        str(user.id)
        for user in instrument.responsible_users
    ]

    if not instrument_form.is_submitted():
        instrument_form.is_markdown.data = (instrument.description_as_html is not None)
        instrument_form.notes_are_markdown.data = (instrument.notes_as_html is not None)
        instrument_form.users_can_create_log_entries.data = instrument.users_can_create_log_entries
        instrument_form.users_can_view_log_entries.data = instrument.users_can_view_log_entries
        instrument_form.create_log_entry_default.data = instrument.create_log_entry_default
        instrument_form.is_hidden.data = instrument.is_hidden
    if instrument_form.validate_on_submit():
        if instrument_form.is_markdown.data:
            description_as_html = markdown_to_safe_html(instrument_form.description.data)
            mark_referenced_markdown_images_as_permanent(description_as_html)
        else:
            description_as_html = None

        if instrument_form.notes_are_markdown.data:
            notes_as_html = markdown_to_safe_html(instrument_form.notes.data)
            mark_referenced_markdown_images_as_permanent(notes_as_html)
        else:
            notes_as_html = None

        update_instrument(
            instrument.id,
            instrument_form.name.data,
            instrument_form.description.data,
            description_as_html=description_as_html,
            notes=instrument_form.notes.data,
            notes_as_html=notes_as_html,
            users_can_create_log_entries=instrument_form.users_can_create_log_entries.data,
            users_can_view_log_entries=instrument_form.users_can_view_log_entries.data,
            create_log_entry_default=instrument_form.create_log_entry_default.data,
            is_hidden=instrument_form.is_hidden.data
        )
        flask.flash('The instrument was updated successfully.', 'success')
        instrument_responsible_user_ids = [
            int(user_id)
            for user_id in instrument_form.instrument_responsible_users.data
        ]
        set_instrument_responsible_users(instrument.id, instrument_responsible_user_ids)
        try:
            category_data = json.loads(instrument_form.categories.data)
        except Exception:
            pass
        else:
            themes_by_name = {
                theme.name.lower(): theme
                for theme in list(InstrumentLogCategoryTheme)
            }
            category_keys = {'id', 'title', 'theme'}
            if not isinstance(category_data, list):
                category_data = ()
            existing_category_ids = {
                category.id
                for category in get_instrument_log_categories(instrument_id)
            }
            valid_category_ids = set()
            for category in category_data:
                # skip any invalid entries
                if not isinstance(category, dict):
                    continue
                if set(category.keys()) != category_keys:
                    continue
                if not all(isinstance(category[key], str) for key in category_keys):
                    continue
                category_title = category['title'].strip()
                if not category_title:
                    continue
                if len(category_title) > 100:
                    continue
                try:
                    category_id = int(category['id'])
                except ValueError:
                    continue
                category_theme = themes_by_name.get(category['theme'], InstrumentLogCategoryTheme.GRAY)
                if category_id < 0:
                    new_category = create_instrument_log_category(
                        instrument.id,
                        category['title'],
                        category_theme
                    )
                    valid_category_ids.add(new_category.id)
                elif category_id in existing_category_ids:
                    update_instrument_log_category(
                        category_id,
                        category['title'],
                        category_theme
                    )
                    valid_category_ids.add(category_id)
            for category in get_instrument_log_categories(instrument_id):
                if category.id not in valid_category_ids:
                    delete_instrument_log_category(category.id)
        return flask.redirect(flask.url_for('.instrument', instrument_id=instrument.id))
    return flask.render_template(
        'instruments/instrument_form.html',
        submit_text='Update Instrument',
        instrument_log_category_themes=sorted(InstrumentLogCategoryTheme, key=lambda t: t.value),
        instrument_log_categories=get_instrument_log_categories(instrument.id),
        instrument_form=instrument_form
    )


@frontend.route('/instruments/<int:instrument_id>/log/<int:log_entry_id>/file_attachments/<int:file_attachment_id>')
@flask_login.login_required
def instrument_log_file_attachment(instrument_id, log_entry_id, file_attachment_id):
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)
    if not instrument.users_can_view_log_entries and flask_login.current_user not in instrument.responsible_users:
        return flask.abort(403)
    try:
        file_attachment = get_instrument_log_file_attachment(file_attachment_id)
    except InstrumentLogFileAttachmentDoesNotExistError:
        return flask.abort(404)
    file_extension = os.path.splitext(file_attachment.file_name)[1]
    download = 'preview' not in flask.request.args
    mime_type = flask.current_app.config.get('MIME_TYPES', {}).get(file_extension, '')
    if not mime_type:
        download = True
    if download:
        return flask.Response(
            file_attachment.content,
            200,
            headers={
                'Content-Disposition': f'attachment; filename="{file_attachment.file_name}"'
            }
        )
    else:
        return flask.Response(
            file_attachment.content,
            200,
            headers={
                'Content-Type': mime_type
            }
        )


@frontend.route('/instruments/<int:instrument_id>/log/mobile_upload/<token>', methods=['GET'])
def instrument_log_mobile_file_upload(instrument_id: int, token: str):
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)
    serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='instrument-log-mobile-upload')
    try:
        user_id, instrument_id = serializer.loads(token, max_age=15 * 60)
    except itsdangerous.BadSignature:
        return flask.abort(400)
    try:
        user = get_user(user_id)
    except UserDoesNotExistError:
        return flask.abort(403)
    if not instrument.users_can_create_log_entries and user not in instrument.responsible_users:
        return flask.abort(403)
    if user.is_readonly:
        return flask.abort(403)
    return flask.render_template('mobile_upload.html', user_id=get_user(user_id), instrument=instrument)


@frontend.route('/instruments/<int:instrument_id>/log/mobile_upload/<token>', methods=['POST'])
def post_instrument_log_mobile_file_upload(instrument_id: int, token: str):
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)
    serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='instrument-log-mobile-upload')
    try:
        user_id, instrument_id = serializer.loads(token, max_age=15 * 60)
    except itsdangerous.BadSignature:
        return flask.abort(400)
    try:
        user = get_user(user_id)
    except UserDoesNotExistError:
        return flask.abort(403)
    if not instrument.users_can_create_log_entries and user not in instrument.responsible_users:
        return flask.abort(403)
    if user.is_readonly:
        return flask.abort(403)
    files = flask.request.files.getlist('file_input')
    if not files:
        return flask.redirect(
            flask.url_for(
                '.instrument_log_mobile_file_upload',
                instrument_id=instrument_id,
                token=token
            )
        )
    if files:
        category_ids = []
        log_entry = create_instrument_log_entry(
            instrument_id=instrument_id,
            user_id=user_id,
            content='',
            category_ids=category_ids
        )
        for file_storage in files:
            create_instrument_log_file_attachment(
                instrument_log_entry_id=log_entry.id,
                file_name=file_storage.filename,
                content=file_storage.stream.read()
            )
    return flask.render_template('mobile_upload_success.html')


@flask_login.login_required
@frontend.route('/users/me/settings/instrument_log_order', methods=['POST'])
def set_instrument_log_order():
    form = InstrumentLogOrderForm()
    if not form.validate_on_submit():
        return flask.abort(400)
    ascending = form.ascending.data
    set_user_settings(flask_login.current_user.id, {
        'INSTRUMENT_LOG_ORDER_ASCENDING': ascending
    })
    return "", 200



# coding: utf-8
"""

"""

from .. import db
from .object_permissions import Permissions
from .groups import Group
from .users import User


class Project(db.Model):
    __tablename__ = 'projects'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, unique=True, nullable=False)
    description = db.Column(db.String, nullable=False, default='')


class UserProjectPermissions(db.Model):
    __tablename__ = 'user_project_permissions'

    project_id = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(project_id, user_id),
        {},
    )


class GroupProjectPermissions(db.Model):
    __tablename__ = 'group_project_permissions'

    project_id = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    group_id = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    permissions = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    __table_args__ = (
        db.PrimaryKeyConstraint(project_id, group_id),
        {},
    )


class SubprojectRelationship(db.Model):
    __tablename__ = 'subproject_relationship'

    parent_project_id = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    child_project_id = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    child_can_add_users_to_parent = db.Column(db.Boolean, nullable=False, default=False)

    __table_args__ = (
        db.PrimaryKeyConstraint(parent_project_id, child_project_id),
        {},
    )


class ProjectInvitation(db.Model):
    __tablename__ = 'project_invitations'

    id = db.Column(db.Integer, primary_key=True)
    project_id = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    inviter_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    utc_datetime = db.Column(db.DateTime, nullable=False)
    accepted = db.Column(db.Boolean, nullable=False, default=False)



# coding: utf-8
"""
Script for creating an action in SampleDB.

Usage: python -m sampledb create_action <instrument_id> <type: sample or measurement > <name> <description> <schema_file_name>
"""

import json
import sys
from .. import create_app
from ..logic.actions import create_action, get_action_type, get_action_types
from ..logic.instruments import get_instrument
from ..logic.schemas import validate_schema
from ..logic.errors import InstrumentDoesNotExistError, ValidationError, ActionTypeDoesNotExistError
from .. import models


def main(arguments):
    if len(arguments) != 5:
        print(__doc__)
        exit(1)
    instrument_id, action_type_id, name, description, schema_file_name = arguments
    if instrument_id == 'None':
        instrument_id = None
    else:
        try:
            instrument_id = int(instrument_id)
        except ValueError:
            print("Error: instrument_id must be an integer or 'None'", file=sys.stderr)
            exit(1)
    try:
        action_type_id = int(action_type_id)
    except ValueError:
        action_type_id = {
            'sample': models.ActionType.SAMPLE_CREATION,
            'measurement': models.ActionType.MEASUREMENT,
            'simulation': models.ActionType.SIMULATION
        }.get(action_type_id, None)
    try:
        get_action_type(action_type_id)
    except ActionTypeDoesNotExistError:
        action_type_id = None
    if action_type_id is None:
        print("Error: action type must be one of the following:", file=sys.stderr)
        print(f'- "sample" (for {get_action_type(models.ActionType.SAMPLE_CREATION).name})', file=sys.stderr)
        print(f'- "measurement" (for {get_action_type(models.ActionType.MEASUREMENT).name})', file=sys.stderr)
        print(f'- "simulation" (for {get_action_type(models.ActionType.SIMULATION).name})', file=sys.stderr)
        for action_type in get_action_types():
            print(f'- "{action_type.id}" (for {action_type.name})', file=sys.stderr)
        exit(1)
    app = create_app()
    with app.app_context():
        if instrument_id is not None:
            try:
                get_instrument(instrument_id)
            except InstrumentDoesNotExistError:
                print('Error: no instrument with this id exists', file=sys.stderr)
                exit(1)
        with open(schema_file_name, 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        try:
            validate_schema(schema)
        except ValidationError as e:
            print('Error: invalid schema: {}'.format(str(e)), file=sys.stderr)
            exit(1)
        action = create_action(
            instrument_id=instrument_id,
            action_type_id=action_type_id,
            name=name,
            description=description,
            schema=schema
        )
        print("Success: the action has been created in SampleDB (#{})".format(action.id))



# coding: utf-8
"""

"""

import flask
import flask_login

from .. import frontend

from ...logic.api_log import get_api_log_entries
from ...logic.authentication import Authentication, AuthenticationType

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


@frontend.route('/users/me/api_token_id/<int:api_token_id>/log/')
@flask_login.login_required
def current_user_api_log(api_token_id):
    return flask.redirect(flask.url_for('.api_log', user_id=flask_login.current_user.id, api_token_id=api_token_id))


@frontend.route('/users/<int:user_id>/api_token_id/<int:api_token_id>/log/')
@flask_login.login_required
def api_log(user_id, api_token_id):
    if user_id != flask_login.current_user.id and not flask_login.current_user.is_admin:
        return flask.abort(404)
    api_token = Authentication.query.filter_by(
        id=api_token_id,
        user_id=user_id,
        type=AuthenticationType.API_TOKEN
    ).first()
    if api_token is None:
        return flask.abort(404)
    api_log_entries = get_api_log_entries(api_token.id)
    return flask.render_template(
        'api_log.html',
        api_log_entries=api_log_entries,
        api_token_description=api_token.login.get('description', '-')
    )



# coding: utf-8
"""
RESTful API for SampleDB
"""

from flask_restful import Resource

from .authentication import multi_auth
from ...logic.instruments import get_instrument, get_instruments
from ...logic import errors

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def instrument_to_json(instrument):
    return {
        'instrument_id': instrument.id,
        'name': instrument.name,
        'description': instrument.description,
        'is_hidden': instrument.is_hidden,
        'instrument_scientists': [user.id for user in instrument.responsible_users]
    }


class Instrument(Resource):
    @multi_auth.login_required
    def get(self, instrument_id: int):
        try:
            instrument = get_instrument(instrument_id=instrument_id)
        except errors.InstrumentDoesNotExistError:
            return {
                "message": "instrument {} does not exist".format(instrument_id)
            }, 404
        return instrument_to_json(instrument)


class Instruments(Resource):
    @multi_auth.login_required
    def get(self):
        instruments = get_instruments()
        return [instrument_to_json(instrument) for instrument in instruments]



# coding: utf-8
"""
Scripts for administrating SampleDB

Usage: python -m sampledb <script> ...
"""

import sys

from .scripts import script_documentation, script_functions


available_scripts = list(script_functions.keys())
available_scripts.sort()


def help_and_exit(return_code):
    print(
        __doc__ + '\n'
        'Available scripts:\n'
        ' - ' + '\n - '.join(available_scripts) + '\n\n'
        'To find out how more about a script, use:\n\n'
        'python -m sampledb help <script>\n'
    )
    exit(return_code)


def main(argv):
    if len(argv) < 2:
        return help_and_exit(1)
    script = argv[1]
    arguments = argv[2:]
    if script in ('help', '--help', '-h'):
        if arguments and arguments[0] in script_documentation:
            print(script_documentation[arguments[0]])
            exit(0)
        return help_and_exit(0)
    if any(argument in ('help', '--help', '-h') for argument in arguments):
        if script in script_documentation:
            print(script_documentation[script])
            exit(0)
        return help_and_exit(0)
    if script not in script_functions:
        return help_and_exit(2)
    script_functions[script](arguments)


if __name__ == '__main__':
    main(sys.argv)



# coding: utf-8
"""

"""

import enum
import datetime
import typing

from .. import db

from .users import User


@enum.unique
class NotificationType(enum.Enum):
    OTHER = 0
    ASSIGNED_AS_RESPONSIBLE_USER = 1
    INVITED_TO_GROUP = 2
    INVITED_TO_PROJECT = 3
    ANNOUNCEMENT = 4
    RECEIVED_OBJECT_PERMISSIONS_REQUEST = 5
    INSTRUMENT_LOG_ENTRY_CREATED = 6
    REFERENCED_BY_OBJECT_METADATA = 7
    INSTRUMENT_LOG_ENTRY_EDITED = 8


@enum.unique
class NotificationMode(enum.Enum):
    IGNORE = 0
    WEBAPP = 1
    EMAIL = 2


class Notification(db.Model):
    __tablename__ = 'notifications'

    id = db.Column(db.Integer, primary_key=True)
    type = db.Column(db.Enum(NotificationType), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    data = db.Column(db.JSON, nullable=False)
    was_read = db.Column(db.Boolean, nullable=False)
    utc_datetime = db.Column(db.DateTime, nullable=False)

    def __init__(self, type: NotificationType, user_id: int, data: typing.Dict[str, typing.Any], utc_datetime: typing.Optional[datetime.datetime] = None):
        self.type = type
        self.user_id = user_id
        self.data = data
        self.was_read = False
        if utc_datetime is None:
            utc_datetime = datetime.datetime.utcnow()
        self.utc_datetime = utc_datetime

    def __repr__(self):
        return '<{0}(id={1.id}, type={1.type}, data={1.data})>'.format(type(self).__name__, self)


class NotificationModeForType(db.Model):
    __tablename__ = 'notification_mode_for_types'

    id = db.Column(db.Integer, primary_key=True)
    type = db.Column(db.Enum(NotificationType), nullable=True)
    user_id = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    mode = db.Column(db.Enum(NotificationMode), nullable=False)
    __table_args__ = (
        db.UniqueConstraint('type', 'user_id', name='_notification_mode_for_types_uc'),
    )

    def __init__(self, type: typing.Optional[NotificationType], user_id: int, mode: NotificationMode):
        self.type = type
        self.user_id = user_id
        self.mode = mode

    def __repr__(self):
        return '<{0}(type={1.type}, user_id={1.user_id}, mode={1.mode})>'.format(type(self).__name__, self)



# coding: utf-8
"""
Add responsible_user_id column to object location assignments table and make location_id nullable.
"""

import os

MIGRATION_INDEX = 6
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    # Skip migration by condition
    column_names = db.session.execute("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'object_location_assignments'
    """).fetchall()
    if ('responsible_user_id',) in column_names:
        return False

    # Perform migration
    db.session.execute("""
        ALTER TABLE object_location_assignments
        ADD responsible_user_id INTEGER
    """)
    db.session.execute("""
        ALTER TABLE object_location_assignments
        ADD FOREIGN KEY(responsible_user_id) REFERENCES users(id)
    """)
    db.session.execute("""
        ALTER TABLE object_location_assignments
        ALTER COLUMN location_id DROP NOT NULL
    """)
    return True



# coding: utf-8
"""

"""

import typing

from .. import db


class ActionType(db.Model):
    __tablename__ = 'action_types'

    # default action type IDs
    # offset to -100 to allow later addition of new default action types
    # see: migrations/action_type_create_default_action_types.py
    SAMPLE_CREATION = -100 + 1
    MEASUREMENT = -100 + 2
    SIMULATION = -100 + 3

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, nullable=False)
    description = db.Column(db.String, nullable=False)
    object_name = db.Column(db.String, nullable=False)
    object_name_plural = db.Column(db.String, nullable=False)
    view_text = db.Column(db.String, nullable=False)
    perform_text = db.Column(db.String, nullable=False)
    admin_only = db.Column(db.Boolean, nullable=False, default=False)
    show_on_frontpage = db.Column(db.Boolean, nullable=False, default=False)
    show_in_navbar = db.Column(db.Boolean, nullable=False, default=False)
    enable_labels = db.Column(db.Boolean, nullable=False, default=True)
    enable_files = db.Column(db.Boolean, nullable=False, default=True)
    enable_locations = db.Column(db.Boolean, nullable=False, default=True)
    enable_publications = db.Column(db.Boolean, nullable=False, default=True)
    enable_comments = db.Column(db.Boolean, nullable=False, default=True)
    enable_activity_log = db.Column(db.Boolean, nullable=False, default=True)
    enable_related_objects = db.Column(db.Boolean, nullable=False, default=True)

    def __repr__(self):
        return '<{0}(id={1.id!r}, name={1.name!r})>'.format(type(self).__name__, self)


class Action(db.Model):
    __tablename__ = 'actions'

    id = db.Column(db.Integer, primary_key=True)
    type_id = db.Column(db.Integer, db.ForeignKey("action_types.id"), nullable=False)
    type = db.relationship(ActionType)
    name = db.Column(db.String, nullable=False)
    description = db.Column(db.String, nullable=False, default='')
    instrument_id = db.Column(db.Integer, db.ForeignKey("instruments.id"), nullable=True)
    instrument = db.relationship("Instrument", backref="actions")
    schema = db.Column(db.JSON, nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey("users.id"), nullable=True)
    user = db.relationship("User", backref="actions")
    description_as_html = db.Column(db.String, nullable=True, default=None)
    is_hidden = db.Column(db.Boolean, nullable=False, default=False)

    def __init__(self, action_type_id: int, name: str, schema: dict, description: str = '', instrument_id: typing.Optional[int] = None, user_id: typing.Optional[int] = None, description_as_html: typing.Optional[str] = None, is_hidden: bool = False):
        self.type_id = action_type_id
        self.name = name
        self.description = description
        self.instrument_id = instrument_id
        self.schema = schema
        self.user_id = user_id
        self.description_as_html = description_as_html
        self.is_hidden = is_hidden

    def __eq__(self, other):
        return (
            self.id == other.id and
            self.instrument_id == other.instrument_id and
            self.user_id == other.user_id and
            self.name == other.name and
            self.description == other.description and
            self.description_as_html == other.description_as_html and
            self.is_hidden == other.is_hidden and
            self.schema == other.schema
        )

    def __repr__(self):
        return '<{0}(id={1.id!r}, name={1.name!r})>'.format(type(self).__name__, self)



# coding: utf-8
"""
Logic module for management of actions

Actions are used to represent all kinds of methods or processes that result
in the creation of a new sample or measurement. What kind of object is
created when performing an action is defined by the action's type. The
action's schema defines what information should or may be recorded in SampleDB
for a newly created object.

Actions can be related to an instrument, which serves both to group actions
together and to provide instrument responsible users with permissions for
samples or measurements created with their instrument.

Actions can also be user-defined, to allow advanced users to create actions
for instruments which would otherwise not be included.

As actions form the basis for objects, they cannot be deleted. However, an
action can be altered as long as the type and instrument stay the same.
"""

import collections
import typing

from .. import db
from .. import models
from ..models import Action
from . import errors, instruments, users, schemas


class ActionType(collections.namedtuple('ActionType', [
    'id',
    'name',
    'description',
    'object_name',
    'object_name_plural',
    'view_text',
    'perform_text',
    'admin_only',
    'show_on_frontpage',
    'show_in_navbar',
    'enable_labels',
    'enable_files',
    'enable_locations',
    'enable_publications',
    'enable_comments',
    'enable_activity_log',
    'enable_related_objects'
])):
    """
    This class provides an immutable wrapper around models.actions.ActionType.
    """

    def __new__(cls, id: int, name: str, description: str, object_name: str, object_name_plural: str, view_text: str, perform_text: str, admin_only: bool, show_on_frontpage: bool, show_in_navbar: bool, enable_labels: bool, enable_files: bool, enable_locations: bool, enable_publications: bool, enable_comments: bool, enable_activity_log: bool, enable_related_objects: bool):
        self = super(ActionType, cls).__new__(cls, id, name, description, object_name, object_name_plural, view_text, perform_text, admin_only, show_on_frontpage, show_in_navbar, enable_labels, enable_files, enable_locations, enable_publications, enable_comments, enable_activity_log, enable_related_objects)
        return self

    @classmethod
    def from_database(cls, action_type: models.ActionType) -> 'ActionType':
        return ActionType(
            id=action_type.id,
            name=action_type.name,
            description=action_type.description,
            object_name=action_type.object_name,
            object_name_plural=action_type.object_name_plural,
            view_text=action_type.view_text,
            perform_text=action_type.perform_text,
            admin_only=action_type.admin_only,
            show_on_frontpage=action_type.show_on_frontpage,
            show_in_navbar=action_type.show_in_navbar,
            enable_labels=action_type.enable_labels,
            enable_files=action_type.enable_files,
            enable_locations=action_type.enable_locations,
            enable_publications=action_type.enable_publications,
            enable_comments=action_type.enable_comments,
            enable_activity_log=action_type.enable_activity_log,
            enable_related_objects=action_type.enable_related_objects
        )

    def __repr__(self):
        return f"<{type(self).__name__}(id={self.id!r}, name={self.name!r})>"


def get_action_types() -> typing.List[ActionType]:
    """
    Return the list of all existing action types.

    :return: the action types
    """
    return [
        ActionType.from_database(action_type)
        for action_type in models.ActionType.query.order_by(models.ActionType.id).all()
    ]


def get_action_type(action_type_id: int) -> ActionType:
    """
    Return the action type with the given action type ID.

    :param action_type_id: the ID of an existing action type
    :return: the action type
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    """
    action_type = models.ActionType.query.get(action_type_id)
    if action_type is None:
        raise errors.ActionTypeDoesNotExistError()
    return ActionType.from_database(action_type)


def create_action_type(
        name: str,
        description: str,
        object_name: str,
        object_name_plural: str,
        view_text: str,
        perform_text: str,
        admin_only: bool,
        show_on_frontpage: bool,
        show_in_navbar: bool,
        enable_labels: bool,
        enable_files: bool,
        enable_locations: bool,
        enable_publications: bool,
        enable_comments: bool,
        enable_activity_log: bool,
        enable_related_objects: bool
) -> ActionType:
    """
    Create a new action type.

    :param name: the name of the action type
    :param description: the description of the action type
    :param object_name: the name of an object created by an action of this type
    :param object_name_plural: the plural form of the object name
    :param view_text: the text to display for viewing objects created by actions of this type
    :param perform_text: the text to display for performing actions of this type
    :param admin_only: whether actions of this type can only be created by administrators
    :param show_on_frontpage: whether this action type should be shown on the frontpage
    :param show_in_navbar: whether actions of this type should be shown in the navbar
    :param enable_labels: whether labels should be enabled for actions of this type
    :param enable_files: whether file uploads should be enabled for actions of this type
    :param enable_locations: whether locations and responsible users should be enabled for actions of this type
    :param enable_publications: whether publications should be enabled for actions of this type
    :param enable_comments: whether comments should be enabled for actions of this type
    :param enable_activity_log: whether the activity log should be enabled for actions of this type
    :param enable_related_objects: whether showing related objects should be enabled for actions of this type
    :return: the created action type
    """
    action_type = models.ActionType(
        name=name,
        description=description,
        object_name=object_name,
        object_name_plural=object_name_plural,
        view_text=view_text,
        perform_text=perform_text,
        admin_only=admin_only,
        show_on_frontpage=show_on_frontpage,
        show_in_navbar=show_in_navbar,
        enable_labels=enable_labels,
        enable_files=enable_files,
        enable_locations=enable_locations,
        enable_publications=enable_publications,
        enable_comments=enable_comments,
        enable_activity_log=enable_activity_log,
        enable_related_objects=enable_related_objects
    )
    db.session.add(action_type)
    db.session.commit()
    return ActionType.from_database(action_type)


def update_action_type(
        action_type_id: int,
        name: str,
        description: str,
        object_name: str,
        object_name_plural: str,
        view_text: str,
        perform_text: str,
        admin_only: bool,
        show_on_frontpage: bool,
        show_in_navbar: bool,
        enable_labels: bool,
        enable_files: bool,
        enable_locations: bool,
        enable_publications: bool,
        enable_comments: bool,
        enable_activity_log: bool,
        enable_related_objects: bool
) -> ActionType:
    """
    Update an existing action type.

    :param action_type_id: the ID of an existing action type
    :param name: the name of the action type
    :param description: the description of the action type
    :param object_name: the name of an object created by an action of this type
    :param object_name_plural: the plural form of the object name
    :param view_text: the text to display for viewing objects created by actions of this type
    :param perform_text: the text to display for performing actions of this type
    :param admin_only: whether actions of this type can only be created by administrators
    :param show_on_frontpage: whether this action type should be shown on the frontpage
    :param show_in_navbar: whether actions of this type should be shown in the navbar
    :param enable_labels: whether labels should be enabled for actions of this type
    :param enable_files: whether file uploads should be enabled for actions of this type
    :param enable_locations: whether locations and responsible users should be enabled for actions of this type
    :param enable_publications: whether publications should be enabled for actions of this type
    :param enable_comments: whether comments should be enabled for actions of this type
    :param enable_activity_log: whether the activity log should be enabled for actions of this type
    :param enable_related_objects: whether showing related objects should be enabled for actions of this type
    :return: the created action type
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    """
    action_type = models.ActionType.query.get(action_type_id)
    if action_type is None:
        raise errors.ActionTypeDoesNotExistError()
    action_type.name = name
    action_type.description = description
    action_type.object_name = object_name
    action_type.object_name_plural = object_name_plural
    action_type.view_text = view_text
    action_type.perform_text = perform_text
    action_type.admin_only = admin_only
    action_type.show_on_frontpage = show_on_frontpage
    action_type.show_in_navbar = show_in_navbar
    action_type.enable_labels = enable_labels
    action_type.enable_files = enable_files
    action_type.enable_locations = enable_locations
    action_type.enable_publications = enable_publications
    action_type.enable_comments = enable_comments
    action_type.enable_activity_log = enable_activity_log
    action_type.enable_related_objects = enable_related_objects
    db.session.add(action_type)
    db.session.commit()
    return ActionType.from_database(action_type)


def create_action(
        action_type_id: int,
        name: str,
        description: str,
        schema: dict,
        instrument_id: typing.Optional[int] = None,
        user_id: typing.Optional[int] = None,
        description_as_html: typing.Optional[str] = None,
        is_hidden: bool = False
) -> Action:
    """
    Creates a new action with the given type, name, description and schema. If
    instrument_id is not None, the action will belong to the instrument with
    this ID.

    :param action_type_id: the ID of an existing action type
    :param name: the name of the action
    :param description: a (possibly empty) description for the action
    :param schema: the schema for objects created using this action
    :param instrument_id: None or the ID of an existing instrument
    :param user_id: None or the ID of an existing user
    :param description_as_html: None or the description as HTML
    :param is_hidden: None or whether or not the action should be hidden
    :return: the created action
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    :raise errors.SchemaValidationError: when the schema is invalid
    :raise errors.InstrumentDoesNotExistError: when instrument_id is not None
        and no instrument with the given instrument ID exists
    :raise errors.UserDoesNotExistError: when user_id is not None and no user
        with the given user ID exists
    """
    # ensure the action type exists
    get_action_type(action_type_id)

    schemas.validate_schema(schema)
    if instrument_id is not None:
        # ensure that the instrument can be found
        instruments.get_instrument(instrument_id)
    if user_id is not None:
        # ensure that the user can be found
        users.get_user(user_id)

    action = Action(
        action_type_id=action_type_id,
        name=name,
        description=description,
        description_as_html=description_as_html,
        is_hidden=is_hidden,
        schema=schema,
        instrument_id=instrument_id,
        user_id=user_id
    )
    db.session.add(action)
    db.session.commit()
    return action


def get_actions(action_type_id: typing.Optional[int] = None) -> typing.List[Action]:
    """
    Returns all actions, optionally only actions of a given type.

    :param action_type_id: None or the ID of an existing action type
    :return: the list of actions
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    """
    if action_type_id is not None:
        actions = Action.query.filter_by(type_id=action_type_id).all()
        if not actions:
            # ensure the action type exists
            get_action_type(action_type_id=action_type_id)
        return actions
    return Action.query.all()


def get_action(action_id: int) -> Action:
    """
    Returns the action with the given action ID.

    :param action_id: the ID of an existing action
    :return: the action
    :raise errors.ActionDoesNotExistError: when no action with the given
        action ID exists
    """
    action = Action.query.get(action_id)
    if action is None:
        raise errors.ActionDoesNotExistError()
    return action


def update_action(
        action_id: int,
        name: str,
        description: str,
        schema: dict,
        description_as_html: typing.Optional[str] = None,
        is_hidden: typing.Optional[bool] = None
) -> None:
    """
    Updates the action with the given action ID, setting its name, description and schema.

    :param action_id: the ID of an existing action
    :param name: the new name of the action
    :param description: the new (possibly empty) description of the action
    :param schema: the new schema for objects created using this action
    :param description_as_html: None or the description as HTML
    :param is_hidden: None or whether or not the action should be hidden
    :raise errors.SchemaValidationError: when the schema is invalid
    :raise errors.InstrumentDoesNotExistError: when instrument_id is not None
        and no instrument with the given instrument ID exists
    """
    schemas.validate_schema(schema)
    action = Action.query.get(action_id)
    if action is None:
        raise errors.ActionDoesNotExistError()
    action.name = name
    action.description = description
    action.description_as_html = description_as_html
    action.schema = schema
    if is_hidden is not None:
        action.is_hidden = is_hidden
    db.session.add(action)
    db.session.commit()



# coding: utf-8
"""

"""

import secrets

import flask
import flask_login
import sqlalchemy.sql.expression

from ... import db

from .. import frontend
from ..authentication_forms import ChangeUserForm, AuthenticationForm, AuthenticationMethodForm
from ..users_forms import RequestPasswordResetForm, PasswordForm, AuthenticationPasswordForm
from ..objects_forms import ObjectPermissionsForm, ObjectUserPermissionsForm, ObjectGroupPermissionsForm, ObjectProjectPermissionsForm
from .forms import NotificationModeForm, OtherSettingsForm, CreateAPITokenForm

from ...logic import user_log
from ...logic.authentication import add_authentication_method, remove_authentication_method, change_password_in_authentication_method, add_api_token
from ...logic.users import get_user, get_users
from ...logic.utils import send_email_confirmation_email, send_recovery_email
from ...logic.security_tokens import verify_token
from ...logic.object_permissions import Permissions, get_default_permissions_for_users, set_default_permissions_for_user, get_default_permissions_for_groups, set_default_permissions_for_group, get_default_permissions_for_projects, set_default_permissions_for_project, default_is_public, set_default_public
from ...logic.projects import get_user_projects, get_project
from ...logic.groups import get_user_groups, get_group
from ...logic.errors import GroupDoesNotExistError, UserDoesNotExistError, ProjectDoesNotExistError
from ...logic.notifications import NotificationMode, NotificationType, get_notification_modes, set_notification_mode_for_type
from ...logic.settings import get_user_settings, set_user_settings

from ...models import Authentication, AuthenticationType


@frontend.route('/users/me/preferences', methods=['GET', 'POST'])
def user_me_preferences():
    if flask_login.current_user.is_authenticated:
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))
    return email_for_resetting_password()


@frontend.route('/users/<int:user_id>/preferences', methods=['GET', 'POST'])
def user_preferences(user_id):
    if 'token' in flask.request.args:
        token = flask.request.args.get('token')
        data = verify_token(token, salt='password', secret_key=flask.current_app.config['SECRET_KEY'])
        if data is not None:
            return reset_password()
        else:
            # es ist egal, ob eingeloggt oder nicht
            return confirm_email()
    elif flask_login.current_user.is_authenticated:
        if user_id != flask_login.current_user.id:
            return flask.abort(403)
        else:
            if not flask_login.login_fresh():
                # ensure only fresh sessions can edit preferences including passwords and api tokens
                return flask.redirect(flask.url_for('.refresh_sign_in', next=flask.url_for('.user_preferences', user_id=flask_login.current_user.id)))
            # user eingeloggt, change preferences möglich
            user = flask_login.current_user
            return change_preferences(user, user_id)
    else:
        return flask.current_app.login_manager.unauthorized()


def change_preferences(user, user_id):
    api_tokens = Authentication.query.filter(Authentication.user_id == user_id, Authentication.type == AuthenticationType.API_TOKEN).all()
    authentication_methods = Authentication.query.filter(Authentication.user_id == user_id, Authentication.type != AuthenticationType.API_TOKEN).all()
    authentication_method_ids = [authentication_method.id for authentication_method in authentication_methods]
    confirmed_authentication_methods = Authentication.query.filter(Authentication.user_id == user_id, Authentication.confirmed == sqlalchemy.sql.expression.true(), Authentication.type != AuthenticationType.API_TOKEN).count()
    change_user_form = ChangeUserForm()
    authentication_form = AuthenticationForm()
    authentication_method_form = AuthenticationMethodForm()
    authentication_password_form = AuthenticationPasswordForm()

    created_api_token = None
    create_api_token_form = CreateAPITokenForm()

    add_user_permissions_form = ObjectUserPermissionsForm()
    add_group_permissions_form = ObjectGroupPermissionsForm()
    add_project_permissions_form = ObjectProjectPermissionsForm()

    notification_mode_form = NotificationModeForm()

    other_settings_form = OtherSettingsForm()
    user_settings = get_user_settings(flask_login.current_user.id)

    user_permissions = get_default_permissions_for_users(creator_id=flask_login.current_user.id)
    group_permissions = get_default_permissions_for_groups(creator_id=flask_login.current_user.id)
    project_permissions = get_default_permissions_for_projects(creator_id=flask_login.current_user.id)
    public_permissions = Permissions.READ if default_is_public(creator_id=flask_login.current_user.id) else Permissions.NONE
    user_permission_form_data = []
    for user_id, permissions in user_permissions.items():
        if user_id is None:
            continue
        user_permission_form_data.append({'user_id': user_id, 'permissions': permissions.name.lower()})
    group_permission_form_data = []
    for group_id, permissions in group_permissions.items():
        if group_id is None:
            continue
        group_permission_form_data.append({'group_id': group_id, 'permissions': permissions.name.lower()})
    project_permission_form_data = []
    for project_id, permissions in project_permissions.items():
        if project_id is None:
            continue
        project_permission_form_data.append({'project_id': project_id, 'permissions': permissions.name.lower()})
    default_permissions_form = ObjectPermissionsForm(public_permissions=public_permissions.name.lower(), user_permissions=user_permission_form_data, group_permissions=group_permission_form_data, project_permissions=project_permission_form_data)

    users = get_users(exclude_hidden=True)
    users = [user for user in users if user.id not in user_permissions]
    groups = get_user_groups(flask_login.current_user.id)
    groups = [group for group in groups if group.id not in group_permissions]
    projects = get_user_projects(flask_login.current_user.id)
    projects = [project for project in projects if project.id not in project_permissions]

    if 'change' not in flask.request.form:
        if change_user_form.name.data is None or change_user_form.name.data == "":
            change_user_form.name.data = user.name
        if change_user_form.email.data is None or change_user_form.email.data == "":
            change_user_form.email.data = user.email
        if change_user_form.orcid.data is None or change_user_form.orcid.data == "":
            change_user_form.orcid.data = user.orcid
        if change_user_form.affiliation.data is None or change_user_form.affiliation.data == "":
            change_user_form.affiliation.data = user.affiliation

    if 'edit' in flask.request.form and flask.request.form['edit'] == 'Edit':
        if authentication_password_form.validate_on_submit() and authentication_password_form.id.data in authentication_method_ids:
            authentication_method_id = authentication_password_form.id.data
            try:
                change_password_in_authentication_method(authentication_method_id, authentication_password_form.password.data)
                flask.flash("Successfully updated your password.", 'success')
            except Exception as e:
                flask.flash("Failed to change password.", 'error')
                return flask.render_template('preferences.html', user=user, change_user_form=change_user_form,
                                             authentication_password_form=authentication_password_form,
                                             default_permissions_form=default_permissions_form,
                                             add_user_permissions_form=add_user_permissions_form,
                                             add_group_permissions_form=add_group_permissions_form,
                                             notification_mode_form=notification_mode_form,
                                             Permissions=Permissions,
                                             NotificationMode=NotificationMode,
                                             NotificationType=NotificationType,
                                             notification_modes=get_notification_modes(flask_login.current_user.id),
                                             user_settings=user_settings,
                                             other_settings_form=other_settings_form,
                                             get_user=get_user,
                                             users=users,
                                             groups=groups,
                                             get_group=get_group,
                                             user_permissions=user_permissions,
                                             group_permissions=group_permissions,
                                             public_permissions=public_permissions,
                                             authentication_method_form=authentication_method_form,
                                             authentication_form=authentication_form,
                                             create_api_token_form=create_api_token_form,
                                             confirmed_authentication_methods=confirmed_authentication_methods,
                                             authentications=authentication_methods, error=str(e), api_tokens=api_tokens)
            user_log.edit_user_preferences(user_id=user_id)
            return flask.redirect(flask.url_for('frontend.user_me_preferences'))
        else:
            flask.flash("Failed to change the password.", 'error')
    if 'change' in flask.request.form and flask.request.form['change'] == 'Change':
        if change_user_form.validate_on_submit():
            if change_user_form.name.data != user.name:
                u = user
                u.name = str(change_user_form.name.data)
                db.session.add(u)
                db.session.commit()
                user_log.edit_user_preferences(user_id=user_id)
                flask.flash("Successfully updated your user name.", 'success')
            if change_user_form.email.data != user.email:
                # send confirm link
                send_email_confirmation_email(
                    email=change_user_form.email.data,
                    user_id=user.id,
                    salt='edit_profile'
                )
                flask.flash("Please see your email to confirm this change.", 'success')
            if change_user_form.orcid.data != user.orcid or change_user_form.affiliation.data != user.affiliation:
                if change_user_form.orcid.data and change_user_form.orcid.data.strip():
                    orcid = change_user_form.orcid.data.strip()
                else:
                    orcid = None
                if change_user_form.affiliation.data and change_user_form.affiliation.data.strip():
                    affiliation = change_user_form.affiliation.data.strip()
                else:
                    affiliation = None
                change_orcid = (user.orcid != orcid and (orcid is not None or user.orcid is not None))
                change_affiliation = (user.affiliation != affiliation and (affiliation is not None or user.affiliation is not None))
                if change_orcid or change_affiliation:
                    user.orcid = orcid
                    user.affiliation = affiliation
                    db.session.add(user)
                    db.session.commit()
                    user_log.edit_user_preferences(user_id=user_id)
                    if change_orcid and not change_affiliation:
                        flask.flash("Successfully updated your ORCID iD.", 'success')
                    elif not change_orcid and change_affiliation:
                        flask.flash("Successfully updated your affiliation.", 'success')
                    else:
                        flask.flash("Successfully updated your ORCID iD and affiliation.", 'success')
            return flask.redirect(flask.url_for('frontend.user_me_preferences'))
    if 'remove' in flask.request.form and flask.request.form['remove'] == 'Remove':
        authentication_method_id = authentication_method_form.id.data
        if authentication_method_form.validate_on_submit():
            try:
                remove_authentication_method(authentication_method_id)
                flask.flash("Successfully removed the authentication method.", 'success')
            except Exception as e:
                flask.flash("Failed to remove the authentication method.", 'error')
                return flask.render_template('preferences.html', user=user, change_user_form=change_user_form,
                                             authentication_password_form=authentication_password_form,
                                             default_permissions_form=default_permissions_form,
                                             add_user_permissions_form=add_user_permissions_form,
                                             add_group_permissions_form=add_group_permissions_form,
                                             notification_mode_form=notification_mode_form,
                                             Permissions=Permissions,
                                             NotificationMode=NotificationMode,
                                             NotificationType=NotificationType,
                                             notification_modes=get_notification_modes(flask_login.current_user.id),
                                             user_settings=user_settings,
                                             other_settings_form=other_settings_form,
                                             get_user=get_user,
                                             users=users,
                                             groups=groups,
                                             get_group=get_group,
                                             user_permissions=user_permissions,
                                             group_permissions=group_permissions,
                                             public_permissions=public_permissions,
                                             authentication_method_form=authentication_method_form,
                                             authentication_form=authentication_form,
                                             create_api_token_form=create_api_token_form,
                                             confirmed_authentication_methods=confirmed_authentication_methods,
                                             authentications=authentication_methods, error=str(e), api_tokens=api_tokens)
            user_log.edit_user_preferences(user_id=user_id)
            return flask.redirect(flask.url_for('frontend.user_me_preferences'))
    if 'add' in flask.request.form and flask.request.form['add'] == 'Add':
        if authentication_form.validate_on_submit():
            # check, if login already exists
            all_authentication_methods = {
                'E': AuthenticationType.EMAIL,
                'L': AuthenticationType.LDAP,
                'O': AuthenticationType.OTHER
            }
            if authentication_form.authentication_method.data not in all_authentication_methods:
                return flask.abort(400)

            authentication_method = all_authentication_methods[authentication_form.authentication_method.data]
            # check, if additional authentication is correct
            try:
                add_authentication_method(user_id, authentication_form.login.data, authentication_form.password.data, authentication_method)
                flask.flash("Successfully added the authentication method.", 'success')
            except Exception as e:
                flask.flash("Failed to add an authentication method.", 'error')
                return flask.render_template('preferences.html', user=user, change_user_form=change_user_form,
                                             authentication_password_form=authentication_password_form,
                                             default_permissions_form=default_permissions_form,
                                             add_user_permissions_form=add_user_permissions_form,
                                             add_group_permissions_form=add_group_permissions_form,
                                             notification_mode_form=notification_mode_form,
                                             Permissions=Permissions,
                                             NotificationMode=NotificationMode,
                                             NotificationType=NotificationType,
                                             notification_modes=get_notification_modes(flask_login.current_user.id),
                                             user_settings=user_settings,
                                             other_settings_form=other_settings_form,
                                             get_user=get_user,
                                             users=users,
                                             groups=groups,
                                             get_group=get_group,
                                             user_permissions=user_permissions,
                                             group_permissions=group_permissions,
                                             public_permissions=public_permissions,
                                             authentication_method_form=authentication_method_form,
                                             authentication_form=authentication_form,
                                             create_api_token_form=create_api_token_form,
                                             confirmed_authentication_methods=confirmed_authentication_methods,
                                             authentications=authentication_methods, error_add=str(e), api_tokens=api_tokens)
            authentication_methods = Authentication.query.filter(Authentication.user_id == user_id, Authentication.type != AuthenticationType.API_TOKEN).all()
        else:
            flask.flash("Failed to add an authentication method.", 'error')
    if 'create_api_token' in flask.request.form and create_api_token_form.validate_on_submit():
        created_api_token = secrets.token_hex(32)
        description = create_api_token_form.description.data
        try:
            add_api_token(flask_login.current_user.id, created_api_token, description)
            api_tokens = Authentication.query.filter(Authentication.user_id == user_id, Authentication.type == AuthenticationType.API_TOKEN).all()
        except Exception as e:
            flask.flash("Failed to add an API token.", 'error')
            return flask.render_template('preferences.html', user=user, change_user_form=change_user_form,
                                         authentication_password_form=authentication_password_form,
                                         default_permissions_form=default_permissions_form,
                                         add_user_permissions_form=add_user_permissions_form,
                                         add_group_permissions_form=add_group_permissions_form,
                                         notification_mode_form=notification_mode_form,
                                         Permissions=Permissions,
                                         NotificationMode=NotificationMode,
                                         NotificationType=NotificationType,
                                         notification_modes=get_notification_modes(flask_login.current_user.id),
                                         user_settings=user_settings,
                                         other_settings_form=other_settings_form,
                                         get_user=get_user,
                                         users=users,
                                         groups=groups,
                                         get_group=get_group,
                                         user_permissions=user_permissions,
                                         group_permissions=group_permissions,
                                         public_permissions=public_permissions,
                                         authentication_method_form=authentication_method_form,
                                         authentication_form=authentication_form,
                                         create_api_token_form=create_api_token_form,
                                         confirmed_authentication_methods=confirmed_authentication_methods,
                                         authentications=authentication_methods, error_add=str(e), api_tokens=api_tokens)
    if 'edit_user_permissions' in flask.request.form and default_permissions_form.validate_on_submit():
        set_default_public(creator_id=flask_login.current_user.id, is_public=(default_permissions_form.public_permissions.data == 'read'))
        for user_permissions_data in default_permissions_form.user_permissions.data:
            user_id = user_permissions_data['user_id']
            try:
                get_user(user_id)
            except UserDoesNotExistError:
                continue
            permissions = Permissions.from_name(user_permissions_data['permissions'])
            set_default_permissions_for_user(creator_id=flask_login.current_user.id, user_id=user_id, permissions=permissions)
        for group_permissions_data in default_permissions_form.group_permissions.data:
            group_id = group_permissions_data['group_id']
            try:
                get_group(group_id)
            except GroupDoesNotExistError:
                continue
            permissions = Permissions.from_name(group_permissions_data['permissions'])
            set_default_permissions_for_group(creator_id=flask_login.current_user.id, group_id=group_id, permissions=permissions)
        for project_permissions_data in default_permissions_form.project_permissions.data:
            project_id = project_permissions_data['project_id']
            try:
                get_project(project_id)
            except ProjectDoesNotExistError:
                continue
            permissions = Permissions.from_name(project_permissions_data['permissions'])
            set_default_permissions_for_project(creator_id=flask_login.current_user.id, project_id=project_id, permissions=permissions)
        flask.flash("Successfully updated default permissions.", 'success')
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))
    if 'add_user_permissions' in flask.request.form and add_user_permissions_form.validate_on_submit():
        user_id = add_user_permissions_form.user_id.data
        permissions = Permissions.from_name(add_user_permissions_form.permissions.data)
        default_user_permissions = get_default_permissions_for_users(creator_id=flask_login.current_user.id)
        assert permissions in [Permissions.READ, Permissions.WRITE, Permissions.GRANT]
        assert user_id not in default_user_permissions
        set_default_permissions_for_user(creator_id=flask_login.current_user.id, user_id=user_id, permissions=permissions)
        flask.flash("Successfully updated default permissions.", 'success')
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))
    if 'add_group_permissions' in flask.request.form and add_group_permissions_form.validate_on_submit():
        group_id = add_group_permissions_form.group_id.data
        permissions = Permissions.from_name(add_group_permissions_form.permissions.data)
        default_group_permissions = get_default_permissions_for_groups(creator_id=flask_login.current_user.id)
        assert permissions in [Permissions.READ, Permissions.WRITE, Permissions.GRANT]
        assert group_id not in default_group_permissions
        set_default_permissions_for_group(creator_id=flask_login.current_user.id, group_id=group_id, permissions=permissions)
        flask.flash("Successfully updated default permissions.", 'success')
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))
    if 'add_project_permissions' in flask.request.form and add_project_permissions_form.validate_on_submit():
        project_id = add_project_permissions_form.project_id.data
        permissions = Permissions.from_name(add_project_permissions_form.permissions.data)
        default_project_permissions = get_default_permissions_for_projects(creator_id=flask_login.current_user.id)
        assert permissions in [Permissions.READ, Permissions.WRITE, Permissions.GRANT]
        assert project_id not in default_project_permissions
        set_default_permissions_for_project(creator_id=flask_login.current_user.id, project_id=project_id, permissions=permissions)
        flask.flash("Successfully updated default permissions.", 'success')
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))
    if 'edit_notification_settings' in flask.request.form and notification_mode_form.validate_on_submit():
        for notification_type in NotificationType:
            if 'notification_mode_for_type_' + notification_type.name.lower() in flask.request.form:
                notification_mode_text = flask.request.form.get('notification_mode_for_type_' + notification_type.name.lower())
                for notification_mode in [NotificationMode.IGNORE, NotificationMode.WEBAPP, NotificationMode.EMAIL]:
                    if notification_mode_text == notification_mode.name.lower():
                        set_notification_mode_for_type(notification_type, flask_login.current_user.id, notification_mode)
                        break
        flask.flash("Successfully updated your notification settings.", 'success')
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))
    confirmed_authentication_methods = Authentication.query.filter(Authentication.user_id == user_id, Authentication.confirmed == sqlalchemy.sql.expression.true(), Authentication.type != AuthenticationType.API_TOKEN).count()
    if 'edit_other_settings' in flask.request.form and other_settings_form.validate_on_submit():
        use_schema_editor = flask.request.form.get('input-use-schema-editor', 'yes') != 'no'
        modified_settings = {
            'USE_SCHEMA_EDITOR': use_schema_editor
        }

        objects_per_page = flask.request.form.get('input-objects-per-page', '')
        if objects_per_page == 'all':
            modified_settings['OBJECTS_PER_PAGE'] = None
        else:
            try:
                modified_settings['OBJECTS_PER_PAGE'] = int(objects_per_page)
            except ValueError:
                pass

        if flask_login.current_user.is_admin:
            use_admin_permissions = flask.request.form.get('input-use-admin-permissions', 'yes') != 'no'
            modified_settings['USE_ADMIN_PERMISSIONS'] = use_admin_permissions
            show_invitation_log = flask.request.form.get('input-show-invitation-log', 'yes') != 'no'
            modified_settings['SHOW_INVITATION_LOG'] = show_invitation_log

        set_user_settings(flask_login.current_user.id, modified_settings)
        flask.flash("Successfully updated your settings.", 'success')
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))
    return flask.render_template('preferences.html', user=user, change_user_form=change_user_form,
                                 authentication_password_form=authentication_password_form,
                                 default_permissions_form=default_permissions_form,
                                 add_user_permissions_form=add_user_permissions_form,
                                 add_group_permissions_form=add_group_permissions_form,
                                 add_project_permissions_form=add_project_permissions_form,
                                 notification_mode_form=notification_mode_form,
                                 NotificationMode=NotificationMode,
                                 NotificationType=NotificationType,
                                 notification_modes=get_notification_modes(flask_login.current_user.id),
                                 user_settings=user_settings,
                                 other_settings_form=other_settings_form,
                                 Permissions=Permissions,
                                 users=users,
                                 get_user=get_user,
                                 groups=groups,
                                 get_group=get_group,
                                 projects=projects,
                                 get_project=get_project,
                                 user_permissions=user_permissions,
                                 group_permissions=group_permissions,
                                 project_permissions=project_permissions,
                                 public_permissions=public_permissions,
                                 authentication_method_form=authentication_method_form,
                                 authentication_form=authentication_form,
                                 create_api_token_form=create_api_token_form,
                                 created_api_token=created_api_token,
                                 confirmed_authentication_methods=confirmed_authentication_methods,
                                 authentications=authentication_methods, api_tokens=api_tokens)


def confirm_email():
    token = flask.request.args.get('token')
    data1 = verify_token(token, salt='edit_profile', secret_key=flask.current_app.config['SECRET_KEY'])
    data2 = verify_token(token, salt='add_login', secret_key=flask.current_app.config['SECRET_KEY'])
    if data1 is None and data2 is None:
        return flask.abort(404)
    else:
        if data1 is not None:
            data = data1
            salt = 'edit_profile'
        else:
            data = data2
            salt = 'add_login'
        if isinstance(data, list) and len(data) == 2:
            # TODO: remove support for old token data
            email, user_id = data
        elif isinstance(data, dict) and 'email' in data and 'user_id' in data:
            email = data['email']
            user_id = data['user_id']
        else:
            return flask.abort(400)
        if salt == 'edit_profile':
            user = get_user(user_id)
            user.email = email
            db.session.add(user)
        elif salt == 'add_login':
            auth = Authentication.query.filter(Authentication.user_id == user_id,
                                               Authentication.login['login'].astext == email).first()
            auth.confirmed = True
            db.session.add(auth)
        else:
            return flask.abort(400)
        db.session.commit()
        user_log.edit_user_preferences(user_id=user_id)
        return flask.redirect(flask.url_for('.user_preferences', user_id=user_id))


def email_for_resetting_password():
    request_password_reset_form = RequestPasswordResetForm()
    if flask.request.method == "GET":
        #  GET (email dialog )
        return flask.render_template('reset_password_by_email.html',
                                     request_password_reset_form=request_password_reset_form)
    if flask.request.method == "POST":
        has_error = False
        if request_password_reset_form.validate_on_submit():
            email = request_password_reset_form.email.data
            if '@' not in email:
                has_error = True
            else:
                send_recovery_email(email)
                return flask.render_template('recovery_email_send.html',
                                             email=email, has_error=has_error)
        return flask.render_template('reset_password_by_email.html',
                                     request_password_reset_form=request_password_reset_form,
                                     has_error=has_error)


def reset_password():
    token = flask.request.args.get('token')
    authentication_id = verify_token(token, salt='password', secret_key=flask.current_app.config['SECRET_KEY'])
    if not authentication_id:
        return flask.abort(404)
    authentication_method = Authentication.query.filter(Authentication.id == authentication_id).first()
    password_form = PasswordForm()
    has_error = False
    if authentication_method is None:
        return flask.abort(404)
    username = authentication_method.login['login']
    if flask.request.method == "GET":
        # confirmation dialog
        return flask.render_template('password.html', password_form=password_form, has_error=has_error,
                                     user=authentication_method.user, username=username)
    else:
        if password_form.validate_on_submit():
            if change_password_in_authentication_method(authentication_method.id, password_form.password.data):
                flask.flash("Your password has been reset.", 'success')
            else:
                flask.flash("Resetting your password failed. Please try again or contact an administrator.", 'error')
            return flask.redirect(flask.url_for('frontend.index'))
        return flask.render_template('password.html', password_form=password_form, has_error=has_error,
                                     user=authentication_method.user, username=username)
